from pint import Quantity

from efootprint.abstract_modeling_classes.explainable_objects import ExplainableHourlyUsage
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u


def create_ram_or_cpu_need_list(time_interval, value):
    need_list = [value - value] * 24

    for hour in range(time_interval[0][0], time_interval[0][1]):
        need_list[hour] = value

    return need_list


def create_ram_need(hours_in_use, ram: Quantity = 100 * u.GB):
    hour_by_hour_ram_need = [0 * u.GB] * 24
    for time_interval in hours_in_use:
        start, end = time_interval
        for i in range(start, end):
            hour_by_hour_ram_need[i] = ram
    return ExplainableHourlyUsage(hour_by_hour_ram_need, "hour_by_hour_ram_need")


def create_cpu_need(hours_in_use, cpu: Quantity = 1 * u.core):
    hour_by_hour_cpu_need = [0 * u.core] * 24
    for time_interval in hours_in_use:
        start, end = time_interval
        for i in range(start, end):
            hour_by_hour_cpu_need[i] = cpu
    return ExplainableHourlyUsage(hour_by_hour_cpu_need, "hour_by_hour_cpu_need")





from unittest import TestCase
from unittest.mock import MagicMock
import unittest

from efootprint.core.system import System
from efootprint.constants.units import u
from efootprint.abstract_modeling_classes.source_objects import SourceValue


class TestSystem(TestCase):
    def setUp(self):
        self.usage_pattern = MagicMock()
        self.server = MagicMock()
        self.storage = MagicMock()
        self.device_population = MagicMock()
        self.network = MagicMock()

        self.usage_pattern.user_journey.servers = {self.server}
        self.usage_pattern.user_journey.storages = {self.storage}
        self.usage_pattern.device_population = self.device_population
        self.usage_pattern.network = self.network

        self.server.instances_fabrication_footprint = SourceValue(100 * u.kg / u.year)
        self.storage.instances_fabrication_footprint = SourceValue(100 * u.kg / u.year)
        self.device_population.instances_fabrication_footprint = SourceValue(100 * u.kg / u.year)

        self.server.energy_footprint = SourceValue(100 * u.kg / u.year)
        self.storage.energy_footprint = SourceValue(100 * u.kg / u.year)
        self.device_population.energy_footprint = SourceValue(100 * u.kg / u.year)
        self.network.energy_footprint = SourceValue(100 * u.kg / u.year)

        self.system = System(
            "Non cloud system",
            usage_patterns=[self.usage_pattern]
        )

    def test_servers(self):
        self.assertEqual({self.server}, self.system.servers)

    def test_storages(self):
        self.assertEqual({self.storage}, self.system.storages)

    def test_device_populations(self):
        self.assertEqual({self.device_population}, self.system.device_populations)

    def test_networks(self):
        self.assertEqual({self.network}, self.system.networks)

    def test_total_fabrication_footprints(self):
        expected_dict = {
            "Servers": SourceValue(100 * u.kg / u.year),
            "Storage": SourceValue(100 * u.kg / u.year),
            "Devices": SourceValue(100 * u.kg / u.year),
            "Network": SourceValue(0 * u.kg / u.year)
        }
        self.assertDictEqual(expected_dict, self.system.total_fabrication_footprints())

    def test_total_energy_footprints(self):
        energy_footprints = self.system.total_energy_footprints()
        expected_dict = {
            "Servers": SourceValue(100 * u.kg / u.year),
            "Storage": SourceValue(100 * u.kg / u.year),
            "Devices": SourceValue(100 * u.kg / u.year),
            "Network": SourceValue(100 * u.kg / u.year)
        }

        self.assertDictEqual(expected_dict, energy_footprints)


if __name__ == '__main__':
    unittest.main()



from efootprint.constants.units import u
from efootprint.constants.sources import Sources
from efootprint.abstract_modeling_classes.source_objects import SourceValue, SourceObject
from efootprint.core.hardware.device_population import DevicePopulation

from unittest import TestCase
from unittest.mock import MagicMock, patch


class TestDevicePopulation(TestCase):
    def setUp(self):
        country = MagicMock()
        country.average_carbon_intensity = SourceValue(100 * u.g / u.kWh)
        self.device1 = MagicMock()
        self.device1.lifespan = SourceValue(1 * u.year, Sources.HYPOTHESIS)
        self.device1.carbon_footprint_fabrication = SourceValue(10 * u.kg, Sources.BASE_ADEME_V19)
        self.device1.fraction_of_usage_time = SourceValue(2 * u.hour / u.day, Sources.STATE_OF_MOBILE_2022)
        self.device2 = MagicMock()
        self.device2.lifespan = SourceValue(1 * u.year, Sources.HYPOTHESIS)
        self.device2.carbon_footprint_fabrication = SourceValue(10 * u.kg, Sources.BASE_ADEME_V19)
        self.device2.fraction_of_usage_time = SourceValue(2 * u.hour / u.day, Sources.STATE_OF_MOBILE_2022)

        self.device_population = DevicePopulation(
            "Population", SourceValue(2000 * u.user), country, [self.device1, self.device2])
        self.device_population.dont_handle_input_updates = True
        self.usage_pattern = MagicMock()
        self.device_population.modeling_obj_containers = [self.usage_pattern]
        self.usage_pattern.user_journey = MagicMock()
        self.usage_pattern.user_journey_freq_per_user = SourceValue(10 * u.user_journey / (u.user * u.year))
        self.usage_pattern.time_intervals = SourceObject([[8, 16]], Sources.USER_DATA)

    def test_user_journey_freq(self):
        self.device_population.update_user_journey_freq_per_up()
        self.assertEqual(
            20000 * u.user_journey / u.year, self.device_population.user_journey_freq_per_up[self.usage_pattern].value)

    def test_update_nb_user_journeys_in_parallel_during_usage(self):
        with patch.object(self.usage_pattern, "user_journey_freq", SourceValue(2 * u.user_journey / u.year)), \
                patch.object(self.usage_pattern.user_journey, "duration", SourceValue(1 * u.year / u.user_journey)), \
                patch.object(self.usage_pattern, "usage_time_fraction", SourceValue((12 / 24) * u.dimensionless)):
            self.device_population.update_nb_user_journeys_in_parallel_during_usage_per_up()
            self.assertEqual(
                4 * u.user_journey,
                self.device_population.nb_user_journeys_in_parallel_during_usage_per_up[self.usage_pattern].value)

    def test_update_nb_user_journeys_in_parallel_during_usage_round_up(self):
        with patch.object(self.usage_pattern, "user_journey_freq", SourceValue(2 * u.user_journey / u.year)), \
                patch.object(self.usage_pattern.user_journey, "duration",
                             SourceValue(1 * u.year / u.user_journey)), \
                patch.object(self.usage_pattern, "usage_time_fraction",
                             SourceValue((14 / 24) * u.dimensionless)):
            self.device_population.update_nb_user_journeys_in_parallel_during_usage_per_up()
            self.assertEqual(
                4 * u.user_journey,
                self.device_population.nb_user_journeys_in_parallel_during_usage_per_up[self.usage_pattern].value)

    def test_power(self):
        test_device1 = MagicMock()
        test_device1.power = SourceValue(5 * u.W)
        test_device2 = MagicMock()
        test_device2.power = SourceValue(10 * u.W)

        with patch.object(self.device_population, "devices", new=[test_device1, test_device2]), \
             patch.object(self.device_population, "user_journey_freq_per_up",
                {self.usage_pattern: SourceValue(365.25 * u.user_journey / u.year)}), \
             patch.object(self.usage_pattern.user_journey, "duration", SourceValue(1 * u.hour / u.user_journey)):
            self.device_population.update_power()

        self.assertEqual(365.25 * 15 * 1e-3 * u.kWh / u.year, self.device_population.power.value)

    def test_energy_footprint(self):
        with patch.object(
                self.device_population, "user_journey_freq_per_up",
                {self.usage_pattern: SourceValue(365.25 * u.user_journey / u.year)}):
            self.device_population.power = SourceValue(1000 * u.kWh / u.year)
            self.device_population.update_energy_footprint()
            self.assertEqual(100 * u.kg / u.year, self.device_population.energy_footprint.value)

    def test_fabrication_footprint(self):
        with patch.object(
                self.device_population, "user_journey_freq_per_up",
                {self.usage_pattern: SourceValue(365.25 * u.user_journey / u.year)}),\
                patch.object(self.usage_pattern.user_journey, "duration", SourceValue(1 * u.hour / u.user_journey)):
            self.device_population.update_instances_fabrication_footprint()
            self.assertEqual(
                2 * 5 * u.kg / u.year, round(self.device_population.instances_fabrication_footprint.value, 2))



from efootprint.abstract_modeling_classes.explainable_objects import ExplainableHourlyUsage, ExplainableQuantity
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.constants.sources import Sources, SOURCE_VALUE_DEFAULT_NAME
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u

from abc import abstractmethod
from typing import List


class Hardware(ModelingObject):
    def __init__(self, name: str, carbon_footprint_fabrication: SourceValue, power: SourceValue,
                 lifespan: SourceValue, fraction_of_usage_time: SourceValue):
        super().__init__(name)
        self.carbon_footprint_fabrication = carbon_footprint_fabrication.set_label(f"Carbon footprint fabrication of {self.name}")
        self.power = power.set_label(f"Power of {self.name}")
        self.lifespan = lifespan.set_label(f"Lifespan of {self.name}")
        if not fraction_of_usage_time.value.check("[]"):
            raise ValueError("Variable 'fraction_of_usage_per_day' shouldn’t have any dimensionality")
        self.fraction_of_usage_time = fraction_of_usage_time.set_label(f"{self.name} fraction of usage time")

        self.calculated_attributes = []

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List[ModelingObject]:
        return self.modeling_obj_containers


class InfraHardware(Hardware):
    def __init__(self, name: str, carbon_footprint_fabrication: SourceValue, power: SourceValue, lifespan: SourceValue,
                 average_carbon_intensity: SourceValue):
        super().__init__(
            name, carbon_footprint_fabrication, power, lifespan, SourceValue(1 * u.dimensionless, Sources.HYPOTHESIS))
        self.all_services_cpu_needs = None
        self.all_services_ram_needs = None
        self.nb_of_instances = None
        self.instances_power = None
        self.fraction_of_time_in_use = None
        self.energy_footprint = None
        self.instances_fabrication_footprint = None
        if not average_carbon_intensity.value.check("[time]**2 / [length]**2"):
            raise ValueError(
                "Variable 'average_carbon_intensity' does not have mass over energy "
                "('[time]**2 / [length]**2') dimensionality"
            )
        self.average_carbon_intensity = average_carbon_intensity
        if self.average_carbon_intensity.label == SOURCE_VALUE_DEFAULT_NAME:
            self.average_carbon_intensity.set_label(f"Average carbon intensity of {self.name} electricity")

        self.calculated_attributes_defined_in_infra_hardware_class = [
            "all_services_cpu_needs", "all_services_ram_needs", "fraction_of_time_in_use",
            "nb_of_instances", "instances_fabrication_footprint", "instances_power", "energy_footprint"
        ]

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List:
        return []

    @abstractmethod
    def update_nb_of_instances(self):
        pass

    @abstractmethod
    def update_instances_power(self):
        pass

    @property
    def services(self):
        return self.modeling_obj_containers

    def update_all_services_ram_needs(self):
        if len(self.services) > 0:
            service_ram_needs_list = [service.hour_by_hour_ram_need for service in self.services]
            all_service_ram_needs = sum(service_ram_needs_list)

            self.all_services_ram_needs = all_service_ram_needs.set_label(
                f"RAM needs of all services running on {self.name}")
        else:
            self.all_services_ram_needs = ExplainableHourlyUsage(
                [0 * u.GB] * 24,
                f"No RAM need for {self.name} because no associated service")

    def update_all_services_cpu_needs(self):
        if len(self.services) > 0:
            service_cpu_needs_list = [service.hour_by_hour_cpu_need for service in self.services]
            all_services_cpu_needs = sum(service_cpu_needs_list)

            self.all_services_cpu_needs = all_services_cpu_needs.set_label(
                f"CPU needs of all services running on {self.name}")
        else:
            self.all_services_cpu_needs = ExplainableHourlyUsage(
                [0 * u.core] * 24,
                f"No CPU need for {self.name} because no associated service")

    def update_fraction_of_time_in_use(self):
        if len(self.services) > 0:
            hourly_usage_sum = self.all_services_cpu_needs.to_usage() + self.all_services_ram_needs.to_usage()

        else:
            hourly_usage_sum = ExplainableHourlyUsage(
                [0 * u.dimensionless] * 24,
                f"No activity for {self.name} because no associated service")

        fraction_of_time_in_use = hourly_usage_sum.compute_usage_time_fraction()

        self.fraction_of_time_in_use = fraction_of_time_in_use.set_label(
            f"Fraction of time in use of {self.name}")

    def update_instances_fabrication_footprint(self):
        self.instances_fabrication_footprint = (
                self.carbon_footprint_fabrication * self.nb_of_instances / self.lifespan).to(
            u.kg / u.year).set_label(f"Instances of {self.name} fabrication footprint")

    def update_energy_footprint(self):
        self.energy_footprint = (self.instances_power * self.average_carbon_intensity).to(
            u.kg / u.year).set_label(f"Energy footprint of {self.name}")



from efootprint.abstract_modeling_classes.explainable_objects import ExplainableHourlyUsage
from efootprint.constants.countries import Countries
from efootprint.constants.sources import Sources
from efootprint.abstract_modeling_classes.source_objects import SourceValue, SourceObject
from efootprint.core.usage.usage_pattern import UsagePattern
from efootprint.constants.units import u

import unittest
from unittest.mock import MagicMock, patch


class TestUsagePattern(unittest.TestCase):
    def setUp(self):
        self.storage = MagicMock()
        self.server = MagicMock()
        self.service1 = MagicMock()
        self.service2 = MagicMock()

        user_journey = MagicMock()
        user_journey.duration = SourceValue(2.0 * u.min / u.user_journey, label="duration")
        user_journey.data_upload = SourceValue(2.0 * u.MB / u.user_journey, label="data_upload")
        user_journey.data_download = SourceValue(3.0 * u.MB / u.user_journey, label="data_download")

        user_journey.services = [self.service1, self.service2]
        population = MagicMock()
        population.nb_devices = SourceValue(10000 * u.user, label="population")
        population.country = Countries.FRANCE

        network = MagicMock()

        self.usage_pattern = UsagePattern(
            "usage_pattern", user_journey, population, network,
            user_journey_freq_per_user=SourceValue(10 * u.user_journey / (u.user * u.year)),
            time_intervals=SourceObject([[8, 16]], Sources.USER_DATA)
        )

    def test_check_time_intervals_validity(self):
        self.usage_pattern.check_time_intervals_validity([[0, 5], [6, 10], [15, 20]])

    def test_invalid_start_time(self):
        with self.assertRaises(ValueError):
            self.usage_pattern.check_time_intervals_validity([[5, 3], [7, 10]])

    def test_interval_overlap(self):
        with self.assertRaises(ValueError):
            self.usage_pattern.check_time_intervals_validity([[0, 5], [4, 10]])

    def test_update_usage_time_fraction(self):
        hourly_usage = ExplainableHourlyUsage([1 * u.dimensionless] * 2 + [0 * u.dimensionless] * 22, "hourly usage")
        with patch.object(self.usage_pattern, "hourly_usage", hourly_usage):
            self.usage_pattern.update_usage_time_fraction()
            self.assertEqual((2 / 24) * u.dimensionless, self.usage_pattern.usage_time_fraction.value)

    def test_update_hourly_usage(self):
        with patch.object(self.usage_pattern, "time_intervals", SourceObject([[0, 5], [6, 10], [15, 20]])):
            self.usage_pattern.update_hourly_usage()

            for i in range(24):
                if 0 <= i < 5 or 6 <= i < 10 or 15 <= i < 20:
                    self.assertEqual(self.usage_pattern.hourly_usage.value[i], 1 * u.dimensionless)
                else:
                    self.assertEqual(self.usage_pattern.hourly_usage.value[i], 0 * u.dimensionless)

    def test_services(self):
        self.assertEqual([self.service1, self.service2], self.usage_pattern.services)


if __name__ == '__main__':
    unittest.main()



from efootprint.builders.hardware.devices_defaults import default_smartphone, default_laptop, \
    default_box, default_screen
from efootprint.builders.hardware.network_defaults import default_wifi_network, default_mobile_network
from efootprint.builders.hardware.servers_defaults import default_serverless, default_autoscaling, \
    default_onpremise
from efootprint.builders.hardware.storage_defaults import default_ssd, default_hdd

import unittest


class TestDefaultBuilders(unittest.TestCase):
    def test_server_defaults(self):
        serverless = default_serverless("serverless")
        autoscaling = default_autoscaling("autoscaling")
        onpremise = default_onpremise("onpremise")

    def test_storage_defaults(self):
        ssd = default_ssd()
        hdd = default_hdd()

    def test_network_defaults(self):
        wifi = default_wifi_network()
        mobile = default_mobile_network()

    def test_devices_defaults(self):
        smartphone = default_smartphone()
        laptop = default_laptop()
        box = default_box()
        screen = default_screen()



from efootprint.builders.hardware.servers_boaviztapi import print_archetypes_and_their_configs, get_cloud_server, \
    on_premise_server_from_config
from efootprint.constants.sources import Sources
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u

import unittest


class TestBoaviztapiBuilders(unittest.TestCase):
    def test_get_cloud_server(self):
        aws_server = get_cloud_server("aws", "m5.xlarge", SourceValue(100 * u.g / u.kWh, Sources.HYPOTHESIS))

    def test_on_premise_server_from_config(self):
        on_prem_server = on_premise_server_from_config(
            "My server", 2, 24, 6, 16, SourceValue(100 * u.g / u.kWh, Sources.HYPOTHESIS))

    def test_print_archetypes_and_their_configs(self):
        # Too long and not very important so pass for now
        # print_archetypes_and_their_configs()
        pass



from efootprint.constants.sources import Sources
from efootprint.abstract_modeling_classes.source_objects import SourceValue, SourceObject
from efootprint.core.usage.user_journey import UserJourney, UserJourneyStep
from efootprint.core.hardware.servers.autoscaling import Autoscaling
from efootprint.core.hardware.storage import Storage
from efootprint.core.service import Service
from efootprint.core.hardware.device_population import DevicePopulation
from efootprint.core.usage.usage_pattern import UsagePattern
from efootprint.core.hardware.network import Network
from efootprint.core.system import System
from efootprint.constants.countries import Countries
from efootprint.constants.units import u
from efootprint.logger import logger
from efootprint.builders.hardware.devices_defaults import default_laptop
from tests.integration_tests.integration_test_base_class import IntegrationTestBaseClass


class IntegrationTestComplexSystem(IntegrationTestBaseClass):
    @classmethod
    def setUpClass(cls):
        cls.server1 = Autoscaling(
            "Server 1",
            carbon_footprint_fabrication=SourceValue(600 * u.kg, Sources.BASE_ADEME_V19),
            power=SourceValue(300 * u.W, Sources.HYPOTHESIS),
            lifespan=SourceValue(6 * u.year, Sources.HYPOTHESIS),
            idle_power=SourceValue(50 * u.W, Sources.HYPOTHESIS),
            ram=SourceValue(12 * u.GB, Sources.HYPOTHESIS),
            cpu_cores=SourceValue(6 * u.core, Sources.HYPOTHESIS),
            power_usage_effectiveness=SourceValue(1.2 * u.dimensionless, Sources.HYPOTHESIS),
            average_carbon_intensity=SourceValue(100 * u.g / u.kWh, Sources.HYPOTHESIS),
            server_utilization_rate=SourceValue(0.9 * u.dimensionless, Sources.HYPOTHESIS)
        )
        cls.storage = Storage(
            "Default SSD storage",
            carbon_footprint_fabrication=SourceValue(160 * u.kg, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            power=SourceValue(1.3 * u.W, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            lifespan=SourceValue(6 * u.years, Sources.HYPOTHESIS),
            idle_power=SourceValue(0.1 * u.W, Sources.HYPOTHESIS),
            storage_capacity=SourceValue(1 * u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            power_usage_effectiveness=SourceValue(1.2 * u.dimensionless, Sources.HYPOTHESIS),
            average_carbon_intensity=SourceValue(100 * u.g / u.kWh, Sources.HYPOTHESIS),
            data_replication_factor=SourceValue(3 * u.dimensionless, Sources.HYPOTHESIS)
        )
        cls.service1 = Service(
            "Youtube", cls.server1, cls.storage, base_ram_consumption=SourceValue(300 * u.MB, Sources.HYPOTHESIS),
            base_cpu_consumption=SourceValue(2 * u.core, Sources.HYPOTHESIS))
        cls.service2 = Service(
            "Dailymotion", cls.server1, cls.storage, base_ram_consumption=SourceValue(300 * u.MB, Sources.HYPOTHESIS),
            base_cpu_consumption=SourceValue(2 * u.core, Sources.HYPOTHESIS))

        cls.server2 = Autoscaling(
            "Server 2",
            carbon_footprint_fabrication=SourceValue(600 * u.kg, Sources.BASE_ADEME_V19),
            power=SourceValue(300 * u.W, Sources.HYPOTHESIS),
            lifespan=SourceValue(6 * u.year, Sources.HYPOTHESIS),
            idle_power=SourceValue(50 * u.W, Sources.HYPOTHESIS),
            ram=SourceValue(12 * u.GB, Sources.HYPOTHESIS),
            cpu_cores=SourceValue(6 * u.core, Sources.HYPOTHESIS),
            power_usage_effectiveness=SourceValue(1.2 * u.dimensionless, Sources.HYPOTHESIS),
            average_carbon_intensity=SourceValue(100 * u.g / u.kWh, Sources.HYPOTHESIS),
            server_utilization_rate=SourceValue(0.9 * u.dimensionless, Sources.HYPOTHESIS)
        )

        cls.service3 = Service(
            "TikTok server", cls.server2, cls.storage,
            base_ram_consumption=SourceValue(300 * u.MB, Sources.HYPOTHESIS),
            base_cpu_consumption=SourceValue(2 * u.core, Sources.HYPOTHESIS))

        cls.streaming_step = UserJourneyStep(
            "20 min streaming on Youtube", cls.service1, SourceValue(50 * u.kB / u.uj),
            SourceValue((2.5 / 3) * u.GB / u.uj),
            user_time_spent=SourceValue(20 * u.min / u.uj), request_duration=SourceValue(4 * u.min))
        cls.upload_step = UserJourneyStep(
            "0.4s of upload", cls.service1, SourceValue(300 * u.kB / u.uj), SourceValue(0 * u.kB / u.uj),
            user_time_spent=SourceValue(1 * u.s / u.uj), request_duration=SourceValue(0.1 * u.s))
        cls.dailymotion_step = UserJourneyStep(
            "Dailymotion step", cls.service2, SourceValue(300 * u.kB / u.uj), SourceValue(3 * u.MB / u.uj),
            user_time_spent=SourceValue(60 * u.s / u.uj), request_duration=SourceValue(1 * u.s))
        cls.tiktok_step = UserJourneyStep(
            "20 min streaming on Tiktok", cls.service3, SourceValue(50 * u.kB / u.uj),
            SourceValue((2.5 / 3) * u.GB / u.uj),
            user_time_spent=SourceValue(20 * u.min / u.uj), request_duration=SourceValue(4 * u.min))

        cls.uj = UserJourney(
            "Daily video usage", uj_steps=[cls.streaming_step, cls.upload_step, cls.dailymotion_step, cls.tiktok_step])
        cls.device_population = DevicePopulation(
            "French video watchers on laptop", SourceValue(4e7 * 0.3 * u.user), Countries.FRANCE, [default_laptop()])

        cls.network = Network("Default network", SourceValue(0.05 * u("kWh/GB"), Sources.TRAFICOM_STUDY))
        cls.usage_pattern = UsagePattern(
            "Video watching in France", cls.uj, cls.device_population,
            cls.network, SourceValue(365 * u.user_journey / (u.user * u.year)),
            SourceObject([[7, 23]], Sources.USER_DATA))

        cls.system = System("system 1", [cls.usage_pattern])

        cls.initial_footprint = cls.system.total_footprint()
        cls.initial_fab_footprints = {
            cls.storage: cls.storage.instances_fabrication_footprint,
            cls.server1: cls.server1.instances_fabrication_footprint,
            cls.device_population: cls.device_population.instances_fabrication_footprint,
        }
        cls.initial_energy_footprints = {
            cls.storage: cls.storage.energy_footprint,
            cls.server1: cls.server1.energy_footprint,
            cls.network: cls.network.energy_footprint,
            cls.device_population: cls.device_population.energy_footprint,
        }

    def test_remove_service2_uj_step(self):
        logger.warning("Removing service2 uj step")
        self.uj.uj_steps = [self.streaming_step, self.upload_step]

        self.footprint_has_changed([self.server1, self.storage])
        self.assertNotEqual(self.initial_footprint, self.system.total_footprint())

        logger.warning("Putting service2 uj step back")
        self.uj.uj_steps = [self.streaming_step, self.upload_step, self.dailymotion_step, self.tiktok_step]

        self.footprint_has_not_changed([self.server1, self.storage])
        self.assertEqual(self.initial_footprint, self.system.total_footprint())

    def test_add_new_service(self):
        logger.warning("Adding service")
        new_service = Service(
            "new service", self.server1, self.storage, base_ram_consumption=SourceValue(300 * u.MB, Sources.HYPOTHESIS),
            base_cpu_consumption=SourceValue(1 * u.core, Sources.HYPOTHESIS))
        new_uj = UserJourneyStep(
            "new uj step", new_service, SourceValue(300 * u.kB / u.uj), SourceValue(300 * u.kB / u.uj),
            user_time_spent=SourceValue(1 * u.s / u.uj), request_duration=SourceValue(0.1 * u.s))
        self.uj.uj_steps += [new_uj]

        self.footprint_has_changed([self.server1, self.storage])
        self.assertNotEqual(self.initial_footprint, self.system.total_footprint())

        logger.warning("Removing new service")
        self.uj.uj_steps = self.uj.uj_steps[:-1]
        new_uj.self_delete()
        new_service.self_delete()

        self.footprint_has_not_changed([self.server1, self.storage])
        self.assertEqual(self.initial_footprint, self.system.total_footprint())

    def test_plot_footprints_by_category_and_object(self):
        self.system.plot_footprints_by_category_and_object()

    def test_plot_emission_diffs(self):
        raise NotImplementedError



from efootprint.abstract_modeling_classes.explainable_object_base_class import (
    ObjectLinkedToModelingObj, ExplainableObject)

from typing import Type


class ExplainableObjectDict(ObjectLinkedToModelingObj, dict):
    # TODO: optimization opportunity: as such in the case of several values of an ExplainableObjectDict
    # depending on the same SourceValue if the SourceValue is updated the ExplainableObjectDict will be recomputed as
    # many times as it has values depending on the SourceValue.
    def __init__(self):
        super().__init__()
        self.modeling_obj_container = None
        self.attr_name_in_mod_obj_container = None

    def set_modeling_obj_container(self, new_parent_modeling_object: Type["ModelingObject"], attr_name: str):
        if self.modeling_obj_container is not None and new_parent_modeling_object.id != self.modeling_obj_container.id:
            raise ValueError(
                f"An ExplainableObjectDict can’t be attributed to more than one ModelingObject. Here "
                f"{self.label} is trying to be linked to {new_parent_modeling_object.name} but is already linked to "
                f"{self.modeling_obj_container.name}.")
        self.modeling_obj_container = new_parent_modeling_object
        self.attr_name_in_mod_obj_container = attr_name

    def __setitem__(self, key, value: ExplainableObject):
        if not issubclass(type(value), ExplainableObject):
            raise ValueError(f"ExplainableObjectDicts only accept ExplainableObjects as values, received {type(value)}")
        super().__setitem__(key, value)
        value.set_modeling_obj_container(
            new_modeling_obj_container=self.modeling_obj_container, attr_name=self.attr_name_in_mod_obj_container)



from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.countries import Countries
from efootprint.constants.sources import Sources
from efootprint.core.hardware.network import Network
from efootprint.constants.units import u

from unittest import TestCase
from unittest.mock import MagicMock


class TestNetwork(TestCase):
    def setUp(self):
        self.network = Network("Wifi network", SourceValue(0.05 * u("kWh/GB"), Sources.TRAFICOM_STUDY))
        self.network.dont_handle_input_updates = True

        self.usage_pattern = MagicMock()
        self.usage_pattern.user_journey.data_upload = SourceValue(100 * u.MB / u.user_journey, label='data_upload')
        self.usage_pattern.user_journey.data_download = SourceValue(
            200 * u.MB / u.user_journey, label='data_download')
        self.usage_pattern.user_journey_freq = SourceValue(250 * u.user_journey / u.year)
        self.network.modeling_obj_containers = {self.usage_pattern}
        self.usage_pattern.device_population.country = Countries.FRANCE
        self.network_consumption = (
                self.network.bandwidth_energy_intensity * SourceValue(300 * u.MB / u.user_journey)
        ).to(u.Wh / u.user_journey)

    def test_update_data_upload(self):
        self.network.update_data_upload()
        data_upload = (
                self.usage_pattern.user_journey.data_upload * self.usage_pattern.user_journey_freq).to(u.TB/u.year)
        self.assertEqual(data_upload, self.network.data_upload)

    def test_update_data_download(self):
        self.network.update_data_download()
        data_download = (self.usage_pattern.user_journey.data_download * self.usage_pattern.user_journey_freq).to(
            u.TB/u.year)
        self.assertEqual(data_download, self.network.data_download)

    def test_update_energy_footprint(self):
        self.network.update_energy_footprint()
        uj_freq = self.usage_pattern.user_journey_freq
        carbon_intensity = self.usage_pattern.device_population.country.average_carbon_intensity
        network_consumption = self.network_consumption

        energy_footprint = (uj_freq * carbon_intensity * network_consumption).to(u.kg / u.year)
        self.assertEqual(energy_footprint.value, self.network.energy_footprint.value)



from efootprint.core.hardware.hardware_base_classes import InfraHardware
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u

from abc import abstractmethod


class Server(InfraHardware):
    def __init__(self, name: str, carbon_footprint_fabrication: SourceValue, power: SourceValue,
                 lifespan: SourceValue, idle_power: SourceValue, ram: SourceValue, cpu_cores: SourceValue,
                 power_usage_effectiveness: SourceValue, average_carbon_intensity: SourceValue,
                 server_utilization_rate: SourceValue):
        super().__init__(name, carbon_footprint_fabrication, power, lifespan, average_carbon_intensity)
        self.available_cpu_per_instance = None
        self.available_ram_per_instance = None
        self.server_utilization_rate = None
        self.nb_of_instances = None
        self.idle_power = idle_power.set_label(f"Idle power of {self.name}")
        self.ram = ram.set_label(f"RAM of {self.name}")
        self.cpu_cores = cpu_cores.set_label(f"Nb cpus cores of {self.name}")
        self.power_usage_effectiveness = power_usage_effectiveness.set_label(f"PUE of {self.name}")
        self.server_utilization_rate = server_utilization_rate.set_label(f"{self.name} utilization rate")

        self.calculated_attributes = ["available_ram_per_instance", "available_cpu_per_instance"
                                      ] + self.calculated_attributes_defined_in_infra_hardware_class

    def update_available_ram_per_instance(self):
        services_base_ram_consumptions = [service.base_ram_consumption for service in self.services]
        available_ram_per_instance = self.ram * self.server_utilization_rate
        for service_base_resource_consumption in services_base_ram_consumptions:
            available_ram_per_instance -= service_base_resource_consumption
            if available_ram_per_instance.value < 0 * u.B:
                services_resource_need = sum(services_base_ram_consumptions)
                raise ValueError(
                    f"server has available capacity of "
                    f"{(self.ram * self.server_utilization_rate).value} "
                    f" but is asked {services_resource_need.value}")

        self.available_ram_per_instance = available_ram_per_instance.set_label(
            f"Available RAM per {self.name} instance")

    def update_available_cpu_per_instance(self):
        services_base_cpu_consumptions = [service.base_cpu_consumption for service in self.services]
        available_cpu_per_instance = self.cpu_cores * self.server_utilization_rate
        for service_base_resource_consumption in services_base_cpu_consumptions:
            available_cpu_per_instance -= service_base_resource_consumption
            if available_cpu_per_instance.value < 0:
                services_resource_need = sum(services_base_cpu_consumptions)
                raise ValueError(
                    f"server has available capacity of "
                    f"{(self.cpu_cores * self.server_utilization_rate).value} "
                    f" but is asked {services_resource_need.value}")

        self.available_cpu_per_instance = available_cpu_per_instance.set_label(
            f"Available CPU per {self.name} instance")

    @abstractmethod
    def update_nb_of_instances(self):
        pass

    @abstractmethod
    def update_instances_power(self):
        pass



from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.core.hardware.network import Network
from efootprint.logger import logger

from typing import List
from unittest import TestCase


class IntegrationTestBaseClass(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.initial_energy_footprints = {}
        cls.initial_fab_footprints = {}

    def footprint_has_changed(self, objects_to_test: List[ModelingObject]):
        for obj in objects_to_test:
            try:
                initial_energy_footprint = round(self.initial_energy_footprints[obj].value, 2)
                self.assertNotEqual(initial_energy_footprint, obj.energy_footprint.value)
                if type(obj) != Network:
                    initial_fab_footprint = round(self.initial_fab_footprints[obj].value, 2)
                    new_footprint = round(obj.instances_fabrication_footprint.value + obj.energy_footprint.value, 2)
                    self.assertNotEqual(
                        initial_fab_footprint + initial_energy_footprint, new_footprint)
                    logger.info(
                        f"{obj.name} footprint has changed from {initial_fab_footprint + initial_energy_footprint}"
                        f" to {new_footprint}")
                else:
                    logger.info(f"{obj.name} footprint has changed from "
                                f"{round(initial_energy_footprint, 2)} to {round(obj.energy_footprint.value, 2)}")
            except AssertionError:
                raise AssertionError(f"Footprint hasn’t changed for {obj.name}")

    def footprint_has_not_changed(self, objects_to_test: List[ModelingObject]):
        for obj in objects_to_test:
            try:
                initial_energy_footprint = self.initial_energy_footprints[obj].value
                if type(obj) != Network:
                    initial_fab_footprint = self.initial_fab_footprints[obj].value
                    self.assertEqual(initial_fab_footprint, obj.instances_fabrication_footprint.value)
                self.assertEqual(initial_energy_footprint, obj.energy_footprint.value)
                logger.info(f"{obj.name} footprint is the same as in setup")
            except AssertionError:
                raise AssertionError(f"Footprint has changed for {obj.name}")



from efootprint.constants.sources import Sources
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.hardware.storage import Storage

from unittest import TestCase
from unittest.mock import MagicMock, patch, PropertyMock


class TestStorage(TestCase):
    def setUp(self):
        self.storage_base = Storage(
            "storage_base",
            carbon_footprint_fabrication=SourceValue(0 * u.kg, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            power=SourceValue(0 * u.W, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            lifespan=SourceValue(0 * u.years, Sources.HYPOTHESIS),
            idle_power=SourceValue(0 * u.W, Sources.HYPOTHESIS),
            storage_capacity=SourceValue(0 * u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            power_usage_effectiveness=SourceValue(0 * u.dimensionless, Sources.HYPOTHESIS),
            average_carbon_intensity=SourceValue(100 * u.g / u.kWh),
            data_replication_factor=SourceValue(0 * u.dimensionless, Sources.HYPOTHESIS)
        )
        self.storage_base.dont_handle_input_updates = True

    def test_update_all_services_storage_needs_single_service(self):
        service1 = MagicMock()
        service1.storage = self.storage_base
        service1.storage_needed = SourceValue(2 * u.TB / u.year)
        service2 = MagicMock()
        service2.storage = self.storage_base
        service2.storage_needed = SourceValue(3 * u.TB / u.year)
        with patch.object(Storage, "services", new_callable=PropertyMock) as services_mock:
            services_mock.return_value = {service1, service2}
            self.storage_base.update_all_services_storage_needs()
            self.assertEqual(5 * u.TB / u.year, self.storage_base.all_services_storage_needs.value)

    def test_active_storage_required(self):
        active_storage_expected = (
                SourceValue(10.1 * u.TB / u.year)
                * SourceValue(1 * u.hour, Sources.HYPOTHESIS, "Time interval during which active storage is considered")
        ).to(u.GB)
        with patch.object(self.storage_base, "all_services_storage_needs", new=SourceValue(10.1 * u.TB / u.year)):
            self.storage_base.update_active_storage_required()
            self.assertEqual(active_storage_expected.value, self.storage_base.active_storage_required.value)

    def test_long_term_storage_required(self):
        with patch.object(self.storage_base, "all_services_storage_needs", SourceValue(10.1 * u.TB / u.year)), \
                patch.object(self.storage_base, "active_storage_required", SourceValue(1.5 * u.TB)), \
                patch.object(self.storage_base, "data_replication_factor", SourceValue(3 * u.dimensionless)), \
                patch.object(self.storage_base, "storage_need_from_previous_year", SourceValue(1 * u.TB)):
            self.storage_base.update_long_term_storage_required()

            self.assertEqual(29.8 * u.TB, round(self.storage_base.long_term_storage_required.value, 1))

    def test_nb_of_active_instances(self):
        with patch.object(self.storage_base, "active_storage_required", SourceValue(500 * u.GB)), \
                patch.object(self.storage_base, "storage_capacity", SourceValue(1 * u.TB)):
            self.storage_base.update_nb_of_active_instances()
            self.assertEqual(0.5, round(self.storage_base.nb_of_active_instances.value, 1))

    def test_nb_of_idle_instances(self):
        with patch.object(self.storage_base, "long_term_storage_required", SourceValue(5 * u.GB)), \
                patch.object(self.storage_base, "storage_capacity", SourceValue(1 * u.GB)):
            self.storage_base.update_nb_of_idle_instances()
            self.assertEqual(5, self.storage_base.nb_of_idle_instances.value)

    def test_nb_of_instances(self):
        with patch.object(self.storage_base, "nb_of_active_instances",
                          SourceValue(3 * u.dimensionless)), \
                patch.object(self.storage_base, "nb_of_idle_instances", SourceValue(2 * u.dimensionless)):
            self.storage_base.update_nb_of_instances()
            self.assertEqual(SourceValue(5 * u.dimensionless), self.storage_base.nb_of_instances)

    def test_instances_power(self):
        with patch.object(self.storage_base, "nb_of_active_instances",
                          SourceValue(10 * u.dimensionless)), \
              patch.object(self.storage_base, "nb_of_idle_instances", SourceValue(1 * u.dimensionless)), \
              patch.object(self.storage_base, "fraction_of_time_in_use", SourceValue(0.5 * u.dimensionless)), \
              patch.object(self.storage_base, "power", SourceValue(1.3 * u.W)), \
              patch.object(self.storage_base, "idle_power", SourceValue(0.3 * u.W)), \
              patch.object(self.storage_base, "power_usage_effectiveness", SourceValue(1.2 * u.dimensionless)):
            self.storage_base.update_instances_power()
            expected_power = SourceValue((15.6 / 2 + 0.36) * u.W).to(u.kWh / u.year)
            self.assertEqual(round(expected_power.value, 0), round(self.storage_base.instances_power.value, 0))



from efootprint.abstract_modeling_classes.explainable_objects import ExplainableQuantity, ExplainableHourlyUsage
from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject

import unittest
from unittest.mock import MagicMock, patch
from pint import UnitRegistry
from datetime import datetime
import pytz

u = UnitRegistry()


class TestExplainableQuantity(unittest.TestCase):
    def setUp(self):
        self.a = ExplainableQuantity(1 * u.W, "1 Watt")
        self.b = ExplainableQuantity(2 * u.W, "2 Watt")
        self.c = self.a + self.b
        self.c.set_label("int calc")
        self.d = self.c + self.b
        self.d.set_label("int calc 2")
        self.e = ExplainableQuantity(3 * u.W, "e")
        for explainable_quantity in (self.a, self.b, self.e):
            explainable_quantity.modeling_obj_container = MagicMock(name="name", id="id")
        self.f = self.a + self.b + self.e

    def test_compute_calculation(self):
        self.assertEqual([self.a, self.b, self.e], self.f.direct_ancestors_with_id)

    def test_init(self):
        self.assertEqual(self.a.value, 1 * u.W)
        self.assertEqual(self.a.label, "1 Watt")
        self.assertEqual(self.a.left_parent, None)
        self.assertEqual(self.a.right_parent, None)
        self.assertEqual(self.a.operator, None)

        self.assertEqual(self.c.value, 3 * u.W)
        self.assertEqual(self.c.label, "int calc")
        self.assertEqual(self.c.left_parent, self.a)
        self.assertEqual(self.c.right_parent, self.b)
        self.assertEqual(self.c.operator, '+')

    def test_operators(self):
        self.assertEqual(self.c.value, 3 * u.W)
        self.assertRaises(ValueError, self.a.__add__, 1)
        self.assertRaises(ValueError, self.a.__gt__, 1)
        self.assertRaises(ValueError, self.a.__lt__, 1)
        self.assertRaises(ValueError, self.a.__eq__, 1)

    def test_to(self):
        self.a.to(u.mW)
        self.assertEqual(self.a.value, 1000 * u.mW)

    def test_magnitude(self):
        self.assertEqual(self.a.magnitude, 1)

    def test_add_with_0(self):
        self.assertEqual(self.a, self.a + 0)

    def test_subtract_0(self):
        self.assertEqual(self.a, self.a - 0)


class TestExplainableHourlyUsage(unittest.TestCase):

    def setUp(self):
        self.usage1 = [1 * u.W] * 24
        self.usage2 = [2 * u.W] * 24
        self.hourly_usage1 = ExplainableHourlyUsage(self.usage1, "Usage 1")
        self.hourly_usage2 = ExplainableHourlyUsage(self.usage2, "Usage 2")
        self.sum_hourly_usage = self.hourly_usage1 + self.hourly_usage2

    def test_init(self):
        self.assertEqual(self.hourly_usage1.label, "Usage 1")
        self.assertEqual(len(self.hourly_usage1.value), 24)

    def test_addition(self):
        result_usage = [3 * u.W] * 24
        for value, expected in zip(self.sum_hourly_usage.value, result_usage):
            self.assertEqual(expected, value)

    def test_subtraction(self):
        result = self.hourly_usage2 - self.hourly_usage1
        for value in result.value:
            self.assertEqual(1 * u.W, value)

    @patch('efootprint.abstract_modeling_classes.explainable_objects.datetime')
    def test_convert_to_utc(self, mock_datetime):
        # Artificially fix datetime to avoid test crashing because of annual time changes.
        mock_datetime.now.return_value = datetime(2023, 10, 1)
        usage = ExplainableHourlyUsage(
            [i * u.dimensionless for i in range(1, 6)], "usage")

        # Let's say the local timezone is 2 hours ahead of UTC
        local_tz_ahead_utc = ExplainableObject(pytz.timezone('Europe/Berlin'), "local timezone ahead UTC")
        local_tz_behind_utc = ExplainableObject(pytz.timezone('America/New_York'), "local timezone behind UTC")

        converted_ahead_utc = usage.convert_to_utc(local_tz_ahead_utc)
        converted_behind_utc = usage.convert_to_utc(local_tz_behind_utc)

        # If Berlin is 2 hours ahead, converting to UTC would result in the array shifted by 2 positions to the left
        self.assertEqual([i * u.dimensionless for i in [3, 4, 5, 1, 2]], converted_ahead_utc.value)
        self.assertEqual([i * u.dimensionless for i in [2,3, 4, 5, 1]], converted_behind_utc.value)

        # Check other attributes of converted ExplainableHourlyUsage
        self.assertEqual("", converted_ahead_utc.label)
        self.assertEqual(usage, converted_ahead_utc.left_parent)
        self.assertEqual(local_tz_ahead_utc, converted_ahead_utc.right_parent)
        self.assertEqual("converted to UTC from", converted_ahead_utc.operator)

    def test_compute_usage_time_fraction(self):
        fraction = self.hourly_usage1.compute_usage_time_fraction()
        self.assertEqual(fraction, ExplainableQuantity(1 * u.dimensionless, "1"))

    def test_to_usage(self):
        usage = self.hourly_usage1.to_usage()
        for i, elt in enumerate(usage.value):
            self.assertEqual(1 * u.dimensionless, elt)

    def test_sum(self):
        summed = self.hourly_usage1.sum()
        self.assertEqual(summed, ExplainableQuantity(24 * u.W, "24 W"))

    def test_max(self):
        maximum = self.hourly_usage1.max()
        self.assertEqual(maximum, ExplainableQuantity(1 * u.W, "1 W"))

    def test_eq_returns_true_when_equal(self):
        self.assertTrue(self.hourly_usage1 == self.hourly_usage1)

    def test_eq_returns_false_when_not_equal(self):
        self.assertFalse(self.hourly_usage1 == self.hourly_usage2)


if __name__ == "__main__":
    unittest.main()



from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject, Source
from efootprint.constants.units import u

import numbers
from datetime import datetime
from typing import Type, List
import pytz
from pint import Quantity


class ExplainableQuantity(ExplainableObject):
    def __init__(
            self, value: Quantity, label: str = None, left_child: Type["ExplainableQuantity"] = None,
            right_child: Type["ExplainableQuantity"] = None, child_operator: str = None, source: Source = None):
        if not isinstance(value, Quantity):
            raise ValueError(
                f"Variable 'value' of type {type(value)} does not correspond to the appropriate 'Quantity' type, "
                "it is indeed mandatory to define a unit"
            )
        super().__init__(value, label, left_child, right_child, child_operator, source)

    def to(self, unit_to_convert_to):
        self.value = self.value.to(unit_to_convert_to)

        return self

    @property
    def magnitude(self):
        return self.value.magnitude

    def compare_with_and_return_max(self, other):
        if issubclass(type(other), ExplainableQuantity):
            if self.value >= other.value:
                return ExplainableQuantity(self.value, left_child=self, right_child=other, child_operator="max")
            else:
                return ExplainableQuantity(other.value, left_child=self, right_child=other, child_operator="max")
        else:
            raise ValueError(f"Can only compare with another ExplainableQuantity, not {type(other)}")

    def __gt__(self, other):
        if issubclass(type(other), ExplainableQuantity):
            return self.value > other.value
        else:
            raise ValueError(f"Can only compare with another ExplainableQuantity, not {type(other)}")

    def __lt__(self, other):
        if issubclass(type(other), ExplainableQuantity):
            return self.value < other.value
        else:
            raise ValueError(f"Can only compare with another ExplainableQuantity, not {type(other)}")

    def __eq__(self, other):
        if issubclass(type(other), ExplainableQuantity):
            return self.value == other.value
        else:
            raise ValueError(f"Can only compare with another ExplainableQuantity, not {type(other)}")

    def __add__(self, other):
        if issubclass(type(other), numbers.Number) and other == 0:
            # summing with sum() adds an implicit 0 as starting value
            return ExplainableQuantity(self.value, left_child=self)
        elif issubclass(type(other), ExplainableQuantity):
            return ExplainableQuantity(self.value + other.value, "", self, other, "+")
        else:
            raise ValueError(f"Can only make operation with another ExplainableQuantity, not with {type(other)}")

    def __sub__(self, other):
        if issubclass(type(other), numbers.Number) and other == 0:
            return ExplainableQuantity(self.value, left_child=self)
        elif issubclass(type(other), ExplainableQuantity):
            return ExplainableQuantity(self.value - other.value, "", self, other, "-")
        else:
            raise ValueError(f"Can only make operation with another ExplainableQuantity, not with {type(other)}")

    def __mul__(self, other):
        if issubclass(type(other), ExplainableQuantity):
            return ExplainableQuantity(self.value * other.value, "", self, other, "*")
        elif issubclass(type(other), ExplainableHourlyUsage):
            return other.__mul__(self)
        else:
            raise ValueError(f"Can only make operation with another ExplainableQuantity, not with {type(other)}")

    def __truediv__(self, other):
        if issubclass(type(other), ExplainableQuantity):
            return ExplainableQuantity(self.value / other.value, "", self, other, "/")
        elif issubclass(type(other), ExplainableHourlyUsage):
            return other.__rtruediv__(self)
        else:
            raise ValueError(f"Can only make operation with another ExplainableQuantity, not with {type(other)}")

    def __radd__(self, other):
        return self.__add__(other)

    def __rsub__(self, other):
        if issubclass(type(other), ExplainableQuantity):
            return ExplainableQuantity(other.value - self.value, "", other, self, "-")
        else:
            raise ValueError(f"Can only make operation with another ExplainableQuantity, not with {type(other)}")

    def __rmul__(self, other):
        return self.__mul__(other)

    def __rtruediv__(self, other):
        if issubclass(type(other), ExplainableQuantity):
            return ExplainableQuantity(other.value / self.value, "", other, self, "/")
        elif issubclass(type(other), ExplainableHourlyUsage):
            return other.__truediv__(self)
        else:
            raise ValueError(f"Can only make operation with another ExplainableQuantity, not with {type(other)}")

    def __round__(self, round_level):
        self.value = round(self.value, round_level)
        return self


class ExplainableHourlyUsage(ExplainableObject):
    def __init__(
            self, value: List[Quantity], label: str = None, left_child: ExplainableObject = None,
            right_child: ExplainableObject = None, child_operator: str = None):
        super().__init__(value, label, left_child, right_child, child_operator)

    def convert_to_utc(self, local_timezone):
        utc_tz = pytz.timezone('UTC')
        current_time = datetime.now()
        time_diff = (local_timezone.value.utcoffset(current_time) - utc_tz.utcoffset(current_time))
        time_diff_in_hours = int(time_diff.total_seconds() / 3600)

        return ExplainableHourlyUsage(
            self.value[time_diff_in_hours:] + self.value[:time_diff_in_hours], "",
            left_child=self, right_child=local_timezone, child_operator="converted to UTC from")

    def compute_usage_time_fraction(self):
        unused_hours = 0
        for elt in self.value:
            if not issubclass(type(elt), Quantity):
                raise ValueError("compute_usage_time_fraction should be called on pint Quantity elements")
            else:
                if elt.magnitude == 0:
                    unused_hours += 1

        return ExplainableQuantity(
            ((24 - unused_hours) / 24) * u.dimensionless, "", self, None, "usage time fraction computation")

    def to_usage(self):
        usage_hours = []
        for i, elt in enumerate(self.value):
            if not issubclass(type(elt), Quantity):
                raise ValueError("to_usage method should be called with pint Quantity elements")
            if elt.magnitude != 0:
                usage_hours.append(1 * u.dimensionless)
            else:
                usage_hours.append(0 * u.dimensionless)

        return ExplainableHourlyUsage(usage_hours, "", left_child=self, child_operator="retrieving usage hours")

    def sum(self):
        return ExplainableQuantity(sum(self.value), left_child=self, child_operator="sum")

    def mean(self):
        return ExplainableQuantity(sum(self.value) / 24, left_child=self, child_operator="mean")

    def max(self):
        return ExplainableQuantity(max(self.value), left_child=self, child_operator="max")

    def __eq__(self, other):
        if issubclass(type(other), ExplainableHourlyUsage):
            if len(self.value) != len(other.value):
                raise ValueError(
                    f"Can only compare ExplainableHourlyUsages with values of same length. Here we are trying to "
                    f"compare {self.value} and {other.value}.")
            return_bool = True
            for i in range(len(self.value)):
                if self.value[i] != other.value[i]:
                    return_bool = False
            return return_bool
        else:
            raise ValueError(f"Can only compare with another ExplainableHourlyUsage, not {type(other)}")

    def __add__(self, other):
        if issubclass(type(other), numbers.Number) and other == 0:
            # summing with sum() adds an implicit 0 as starting value
            return ExplainableHourlyUsage(self.value, left_child=self)
        elif issubclass(type(other), ExplainableHourlyUsage):
            return ExplainableHourlyUsage(
                [elt1 + elt2 for elt1, elt2 in zip(self.value, other.value)], "", self, other, "+")
        else:
            raise ValueError(f"Can only make operation with another ExplainableHourlyUsage, not with {type(other)}")

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if issubclass(type(other), numbers.Number) and other == 0:
            return ExplainableHourlyUsage(self.value, left_child=self)
        elif issubclass(type(other), ExplainableHourlyUsage):
            return ExplainableHourlyUsage(
                [elt1 - elt2 for elt1, elt2 in zip(self.value, other.value)], "", self, other, "-")
        else:
            raise ValueError(f"Can only make operation with another ExplainableHourlyUsage, not with {type(other)}")

    def __rsub__(self, other):
        if issubclass(type(other), ExplainableHourlyUsage):
            return ExplainableHourlyUsage(
                [elt1 - elt2 for elt1, elt2 in zip(other.value, self.value)], "", other, self, "-")
        else:
            raise ValueError(f"Can only make operation with another ExplainableHourlyUsage, not with {type(other)}")

    def __mul__(self, other):
        if issubclass(type(other), ExplainableHourlyUsage):
            raise NotImplementedError
        elif issubclass(type(other), ExplainableQuantity):
            return ExplainableHourlyUsage([other.value * elt for elt in self.value], "", self, other, "*")
        else:
            raise ValueError(
                f"Can only make operation with another ExplainableHourlyUsage or ExplainableQuantity, not with {type(other)}")

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if issubclass(type(other), ExplainableHourlyUsage):
            raise NotImplementedError
        elif issubclass(type(other), ExplainableQuantity):
            return ExplainableHourlyUsage([elt / other.value for elt in self.value], "", self, other, "/")
        else:
            raise ValueError(
                f"Can only make operation with another ExplainableHourlyUsage or ExplainableQuantity, not with {type(other)}")

    def __rtruediv__(self, other):
        if issubclass(type(other), ExplainableHourlyUsage):
            raise NotImplementedError
        elif issubclass(type(other), ExplainableQuantity):
            return ExplainableHourlyUsage([other.value / elt for elt in self.value], "", other, self, "/")
        else:
            raise ValueError(
                f"Can only make operation with another ExplainableHourlyUsage or ExplainableQuantity, not with {type(other)}")



from efootprint.constants.sources import Sources
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.hardware.servers.server_base_class import Server

from unittest import TestCase
from unittest.mock import MagicMock, patch, PropertyMock


class TestServerBaseClass(TestCase):
    def setUp(self):
        class TestServer(Server):
            def __init__(self, name: str, carbon_footprint_fabrication: SourceValue, power: SourceValue,
                         lifespan: SourceValue, idle_power: SourceValue, ram: SourceValue, cpu_cores: SourceValue,
                         power_usage_effectiveness: SourceValue, average_carbon_intensity: SourceValue,
                         server_utilization_rate: SourceValue):
                super().__init__(
                    name, carbon_footprint_fabrication, power, lifespan, idle_power, ram, cpu_cores,
                    power_usage_effectiveness,
                    average_carbon_intensity, server_utilization_rate)

            def update_nb_of_instances(self):
                return SourceValue(10 * u.dimensionless)

            def update_instances_power(self):
                return SourceValue(100 * u.W)

        self.country = MagicMock()
        self.server_base = TestServer(
            "Test server",
            carbon_footprint_fabrication=SourceValue(0 * u.kg, Sources.BASE_ADEME_V19),
            power=SourceValue(0 * u.W, Sources.HYPOTHESIS),
            lifespan=SourceValue(0 * u.year, Sources.HYPOTHESIS),
            idle_power=SourceValue(0 * u.W, Sources.HYPOTHESIS),
            ram=SourceValue(0 * u.GB, Sources.HYPOTHESIS),
            cpu_cores=SourceValue(0 * u.core, Sources.HYPOTHESIS),
            power_usage_effectiveness=SourceValue(0 * u.dimensionless, Sources.HYPOTHESIS),
            average_carbon_intensity=SourceValue(100 * u.g / u.kWh),
            server_utilization_rate=SourceValue(0 * u.dimensionless, Sources.HYPOTHESIS)
        )
        self.server_base.dont_handle_input_updates = True

    def test_available_cpu_per_instance_single_service(self):
        service = MagicMock()
        service.base_cpu_consumption = SourceValue(2 * u.core)
        with patch.object(Server, "services", new_callable=PropertyMock) as mock_service, \
                patch.object(self.server_base, "cpu_cores", SourceValue(24 * u.core)), \
                patch.object(self.server_base, "server_utilization_rate", SourceValue(0.7 * u.dimensionless)):
            mock_service.return_value = {service}
            self.server_base.update_available_cpu_per_instance()
            expected_value = SourceValue((24 * 0.7 - 2) * u.core)

            self.assertEqual(expected_value.value, self.server_base.available_cpu_per_instance.value)

    def test_available_ram_per_instance_multiple_services(self):
        service1 = MagicMock()
        service1.base_ram_consumption = SourceValue(10 * u.GB)
        service2 = MagicMock()
        service2.base_ram_consumption = SourceValue(30 * u.GB)
        with patch.object(Server, "services", new_callable=PropertyMock) as mock_service, \
                patch.object(self.server_base, "ram", SourceValue(128 * u.GB)), \
                patch.object(self.server_base, "server_utilization_rate", SourceValue(0.7 * u.dimensionless)):
            mock_service.return_value = {service1, service2}
            self.server_base.update_available_ram_per_instance()
            expected_value = SourceValue((128 * 0.7 - 10 - 30) * u.GB)

            self.assertEqual(expected_value.value, self.server_base.available_ram_per_instance.value)

    def test_available_cpu_per_instance_multiple_services(self):
        service1 = MagicMock()
        service1.base_cpu_consumption = SourceValue(2 * u.core)
        service2 = MagicMock()
        service2.base_cpu_consumption = SourceValue(1 * u.core)
        with patch.object(Server, "services", new_callable=PropertyMock) as mock_service, \
                patch.object(self.server_base, "cpu_cores", SourceValue(24 * u.core)), \
                patch.object(self.server_base, "server_utilization_rate", SourceValue(0.7 * u.dimensionless)):
            mock_service.return_value = {service1, service2}
            self.server_base.update_available_cpu_per_instance()
            expected_value = SourceValue((24 * 0.7 - 2 - 1) * u.core)

            self.assertEqual(expected_value.value, self.server_base.available_cpu_per_instance.value)

    def test_available_ram_per_instance_should_raise_value_error_when_demand_exceeds_server_capacity(self):
        service = MagicMock()
        service.base_ram_consumption = SourceValue(129 * u.GB)
        with patch.object(Server, "services", new_callable=PropertyMock) as mock_service, \
                patch.object(self.server_base, "ram", SourceValue(128 * u.GB)), \
                patch.object(self.server_base, "server_utilization_rate", SourceValue(0.7 * u.dimensionless)):
            mock_service.return_value = {service}
            with self.assertRaises(ValueError):
                self.server_base.update_available_ram_per_instance()



from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject
from efootprint.constants.units import u

from unittest import TestCase
from unittest.mock import MagicMock


class TestExplainableObjectBaseClass(TestCase):
    def setUp(self) -> None:
        self.a = ExplainableObject(1, "a")
        self.b = ExplainableObject(2, "b")

        self.c = ExplainableObject(3, "c")
        self.c.left_parent = self.a
        self.c.right_parent = self.b
        self.c.operator = "+"

        self.d = ExplainableObject(4, "d")
        self.d.left_parent = self.c
        self.d.right_parent = self.a
        self.d.operator = "+"

        self.e = ExplainableObject(5, "e")
        self.e.left_parent = self.c
        self.e.right_parent = self.b
        self.e.operator = "+"
        self.e.label = None

        self.f = ExplainableObject(6, "f")
        self.f.left_parent = self.e
        self.f.right_parent = self.a
        self.f.operator = "+"

        self.g = ExplainableObject(2, "g")
        self.g.left_parent = self.d
        self.g.operator = "root square"

        self.modeling_obj_container_mock = MagicMock()
        self.modeling_obj_container_mock.id = 1
        self.modeling_obj_container_mock.name = "Model1"

    def test_deepcopy_should_set_modeling_object_to_none(self):
        a = ExplainableObject(1, "a")
        a.modeling_obj_container = "obj"
        from copy import deepcopy
        b = deepcopy(a)

        self.assertEqual("a", b.label)
        self.assertEqual(1, b.value)
        self.assertIsNone(b.modeling_obj_container)

    def test_creation_with_label(self):
        eo = ExplainableObject(value=5, label="Label A")
        self.assertEqual(eo.value, 5)
        self.assertEqual(eo.label, "Label A")
        self.assertIsNone(eo.left_parent)
        self.assertIsNone(eo.right_parent)

    def test_creation_without_label_and_child(self):
        with self.assertRaises(ValueError):
            ExplainableObject(value=5)

    def test_set_modeling_obj_container_without_label(self):
        self.a.label = None
        with self.assertRaises(ValueError):
            self.a.set_modeling_obj_container(self.modeling_obj_container_mock, "attr1")

    def test_set_modeling_obj_container_with_different_modeling_object_for_non_input_should_raise_ValueError(self):
        self.a.left_parent = MagicMock()
        self.a.modeling_obj_container = MagicMock()
        self.a.modeling_obj_container.id = 2
        self.a.modeling_obj_container.name = "Model2"
        with self.assertRaises(ValueError):
            self.a.set_modeling_obj_container(self.modeling_obj_container_mock, "attr1")

    def test_set_modeling_obj_container_success(self):
        self.a.set_modeling_obj_container(self.modeling_obj_container_mock, "attr1")
        self.assertEqual(self.a.modeling_obj_container, self.modeling_obj_container_mock)
        self.assertEqual(self.a.attr_name_in_mod_obj_container, "attr1")

    def test_set_modeling_obj_container_should_trigger_update_direct_children_with_id(self):
        ancestor = MagicMock()
        self.a.direct_ancestors_with_id = [ancestor]
        self.a.set_modeling_obj_container(self.modeling_obj_container_mock, "attr1")
        ancestor.update_direct_children_with_id.assert_called_once_with(direct_child=self.a)

    def test_update_direct_children_with_id_shouldnt_update_list_if_child_already_in_list(self):
        self.a.direct_children_with_id = [self.c]
        self.c.modeling_obj_container = self.modeling_obj_container_mock
        self.a.update_direct_children_with_id(self.c)

        self.assertEqual([self.c], self.a.direct_children_with_id)

    def test_get_all_descendants_with_id(self):
        root = ExplainableObject(0, "root")
        child1 = ExplainableObject(1, "child1")
        child1.modeling_obj_container = MagicMock(id="child1_mod_obj_container")
        child2 = ExplainableObject(2, "child2")
        child2.modeling_obj_container = MagicMock(id="child2_mod_obj_container")
        grandchild1 = ExplainableObject(3, "grandchild1")
        grandchild1.modeling_obj_container = MagicMock(id="child1_mod_obj_container")
        grandchild2 = ExplainableObject(4, "grandchild2")
        grandchild2.modeling_obj_container = MagicMock(id="child2_mod_obj_container")

        root.direct_children_with_id.append(child1)
        root.direct_children_with_id.append(child2)
        child1.direct_children_with_id.append(grandchild1)
        child2.direct_children_with_id.append(grandchild2)

        descendants = root.get_all_descendants_with_id()
        descendants_labels = [descendant.label for descendant in descendants]

        self.assertEqual(len(descendants), 4)
        self.assertListEqual(descendants_labels, ['child1', 'grandchild1', 'child2', 'grandchild2'])

    def test_direct_children(self):
        left_parent = ExplainableObject(value=3, label="Label L")
        right_parent = ExplainableObject(value=4, label="Label R")
        left_parent.modeling_obj_container = MagicMock(name="lc_mod_obj_name", id="lc_mod_obj_id")
        right_parent.modeling_obj_container = MagicMock(name="rc_mod_obj_name", id="rc_mod_obj_id")

        eo = ExplainableObject(value=7, left_parent=left_parent, right_parent=right_parent, label="Parent")
        self.assertEqual([left_parent, right_parent], eo.direct_ancestors_with_id)

    def test_set_label(self):
        eo = ExplainableObject(value=5, label="Label A")
        eo.set_label("Intermediate A")
        self.assertEqual(eo.label, "Intermediate A")

    def test_has_child_property(self):
        left_parent = ExplainableObject(value=3, label="Label L")
        eo_with_child = ExplainableObject(value=7, left_parent=left_parent, label="Parent")
        eo_without_child = ExplainableObject(value=7, label="Parent")
        self.assertTrue(eo_with_child.has_parent)
        self.assertFalse(eo_without_child.has_parent)

    def test_explain_simple_sum(self):
        self.assertEqual("c = a + b = 1 + 2 = 3", self.c.explain(pretty_print=False))

    def test_explain_nested_sum(self):
        self.assertEqual("d = c + a = 3 + 1 = 4", self.d.explain(pretty_print=False))

    def test_explain_should_skip_calculus_element_without_label(self):
        self.assertEqual("f = c + b + a = 3 + 2 + 1 = 6", self.f.explain(pretty_print=False))

    def test_explain_without_right_parent(self):
        self.assertEqual("g = root square of (d) = root square of (4) = 2", self.g.explain(pretty_print=False))

    def test_explain_should_put_right_parenthesis_in_complex_calculations(self):
        self.d.label = None
        self.c.label = None
        h = ExplainableObject(1, None, self.c, self.c, "/")
        i = ExplainableObject(2, None, h, self.g, "*")
        j = ExplainableObject(-1, "k", i, self.c, "-")
        self.assertEqual('k = ((a + b) / (a + b)) * g - (a + b) = ((1 + 2) / (1 + 2)) * 2 - (1 + 2) = -1', j.explain(
            pretty_print=False))

    def test_explain_without_children(self):
        eo = ExplainableObject(value=5, label="Label A")
        result = eo.explain()
        self.assertEqual(result, "Label A = 5")

    def test_compute_explain_nested_tuples(self):
        left_parent = ExplainableObject(value=3, label="Label L")
        right_parent = ExplainableObject(value=4, label="Label R")
        eo = ExplainableObject(value=7, left_parent=left_parent, right_parent=right_parent, label="Parent",
                               operator="+")
        result = eo.compute_explain_nested_tuples()
        self.assertEqual(result, (left_parent, '+', right_parent))

    def test_print_tuple_element_value(self):
        self.assertEqual("5.3 W", ExplainableObject.print_tuple_element_value(5.3456 * u.W))
        self.assertEqual(
            "[1.1 yr, 2.2 yr, 3.3 yr]", ExplainableObject.print_tuple_element_value(
                [1.123 * u.year, 2.234 * u.year, 3.345 * u.year]))
        self.assertEqual("[[1, 2], [3, 5]]", ExplainableObject.print_tuple_element_value([[1, 2], [3, 5]]))

    def test_print_tuple_element(self):
        left_parent = ExplainableObject(value=3, label="Label L")
        right_parent = ExplainableObject(value=4, label="Label R")
        eo = ExplainableObject(value=7, left_parent=left_parent, right_parent=right_parent, label="Parent",
                               operator="+")

        self.assertEqual(eo.print_tuple_element((left_parent, '+', right_parent), False), "Label L + Label R")
        self.assertEqual(eo.print_tuple_element((left_parent, '+', right_parent), True), "3 + 4")

    def test_pretty_print_calculation(self):
        calc_str = "Label A = Label L + Label R = 3 + 4 = 7"
        result = ExplainableObject.pretty_print_calculation(calc_str)
        expected_result = """Label A
=
Label L + Label R
=
3 + 4
=
7"""
        self.assertEqual(expected_result, result)

    def test_set_mod_obj_cont_raises_error_if_value_already_linked_to_another_modeling_obj_container_and_children(self):
        self.a.modeling_obj_container = MagicMock(id="mod obj id")
        new_parent_mod_obj = MagicMock(id="another obj id")
        self.a.left_parent = "non null left child"

        with self.assertRaises(ValueError):
            self.a.set_modeling_obj_container(new_parent_mod_obj, "test_attr_name")



from efootprint.abstract_modeling_classes.explainable_objects import ExplainableQuantity, ExplainableHourlyUsage
from efootprint.constants.sources import Sources
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.service import Service

import unittest
from unittest.mock import MagicMock, patch


class TestService(unittest.TestCase):
    def setUp(self):
        self.server = MagicMock()
        self.storage = MagicMock()
        self.base_ram = SourceValue(4 * u.GB, Sources.HYPOTHESIS)
        self.base_cpu = SourceValue(2 * u.core, Sources.HYPOTHESIS)
        self.service = Service("Test Service", self.server, self.storage, self.base_ram, self.base_cpu)
        self.service.dont_handle_input_updates = True

    def test_service_initialization(self):
        self.assertEqual(self.service.name, "Test Service")
        self.assertEqual(self.service.server, self.server)
        self.assertEqual(self.service.storage, self.storage)
        self.assertEqual(self.service.base_ram_consumption, self.base_ram)
        self.assertEqual(self.service.base_cpu_consumption, self.base_cpu)
        self.assertEqual(ExplainableHourlyUsage([0 * u.GB] * 24, " "), self.service.hour_by_hour_ram_need)
        self.assertEqual(ExplainableHourlyUsage([0 * u.core] * 24, " "), self.service.hour_by_hour_cpu_need)
        self.assertEqual(ExplainableQuantity(0 * u.TB / u.year, " "), self.service.storage_needed)

    def test_service_invalid_ram_consumption(self):
        invalid_ram = SourceValue(4 * u.min)
        with self.assertRaises(ValueError):
            Service("Invalid RAM Service", self.server, self.storage, invalid_ram)

    def test_service_invalid_cpu_consumption(self):
        invalid_cpu = SourceValue(2 * u.min)
        with self.assertRaises(ValueError):
            Service("Invalid CPU Service", self.server, self.storage, self.base_ram, invalid_cpu)

    def test_update_storage_needed(self):
        usage_pattern1 = MagicMock()
        usage_pattern1.user_journey_freq = ExplainableQuantity(10 * u.user_journey / u.day, "uj_freq")
        usage_pattern2 = MagicMock()
        usage_pattern2.user_journey_freq = ExplainableQuantity(100 * u.user_journey / u.day, "uj_freq")
        
        uj_step1 = MagicMock()
        uj_step1.service = self.service
        uj_step1.data_upload = ExplainableQuantity(1 * u.MB / u.user_journey, "data_upload")
        uj_step1.usage_patterns = [usage_pattern1]
        uj_step2 = MagicMock()
        uj_step2.service = self.service
        uj_step2.data_upload = ExplainableQuantity(3 * u.MB / u.user_journey, "data_upload")
        uj_step2.usage_patterns = [usage_pattern2]

        with patch.object(self.service, "modeling_obj_containers", [uj_step1, uj_step2]):
            self.service.update_storage_needed()
            self.assertEqual(
                round((310 * u.MB / u.day).to(u.TB / u.year), 3), round(self.service.storage_needed.value, 3))

    def test_update_hour_by_hour_ram_need(self):
        uj_step = MagicMock()
        uj_step.service = self.service
        uj_step.ram_needed = ExplainableQuantity(1.8 * u.GB, "ram_needed")
        uj_step.request_duration = ExplainableQuantity(10 * u.min, "request_duration")

        uj_step2 = MagicMock()
        uj_step2.service = self.service
        uj_step2.ram_needed = ExplainableQuantity(0.6 * u.GB, "ram_needed")
        uj_step2.request_duration = ExplainableQuantity(10 * u.min, "request_duration")

        usage_pattern = MagicMock()
        usage_pattern.user_journey.duration = ExplainableQuantity(1 * u.hour, "uj_duration")
        usage_pattern.nb_user_journeys_in_parallel_during_usage = ExplainableQuantity(
            10 * u.user_journey, "parallel_uj")
        usage_pattern.utc_time_intervals = ExplainableHourlyUsage(
            [1 * u.dimensionless] * 24, "utc_time_intervals")

        for elt in [uj_step, uj_step2]:
            elt.usage_patterns = [usage_pattern]

        with patch.object(self.service, "modeling_obj_containers", [uj_step, uj_step2]):
            self.service.update_hour_by_hour_ram_need()

            self.assertEqual(
                [4 * u.GB] * 24,
                [round(elt, 2) for elt in self.service.hour_by_hour_ram_need.value])

    def test_update_hour_by_hour_cpu_need(self):
        uj_step = MagicMock()
        uj_step.service = self.service
        uj_step.cpu_needed = ExplainableQuantity(1.8 * u.core, "cpu_needed")
        uj_step.request_duration = ExplainableQuantity(10 * u.min, "request_duration")

        uj_step2 = MagicMock()
        uj_step2.service = self.service
        uj_step2.cpu_needed = ExplainableQuantity(0.6 * u.core, "cpu_needed")
        uj_step2.request_duration = ExplainableQuantity(10 * u.min, "request_duration")

        usage_pattern = MagicMock()
        usage_pattern.user_journey.duration = ExplainableQuantity(1 * u.hour, "uj_duration")
        usage_pattern.nb_user_journeys_in_parallel_during_usage = ExplainableQuantity(
            10 * u.user_journey, "uj_in_parallel")
        usage_pattern.utc_time_intervals = ExplainableHourlyUsage([1 * u.dimensionless] * 24, "utc time intervals")

        for elt in [uj_step, uj_step2]:
            elt.usage_patterns = [usage_pattern]

        with patch.object(self.service, "modeling_obj_containers", [uj_step, uj_step2]):
            self.service.update_hour_by_hour_cpu_need()

            self.assertEqual(
                [4 * u.core] * 24,
                [round(elt, 2) for elt in self.service.hour_by_hour_cpu_need.value])

    def test_self_delete_should_raise_error_if_self_has_associated_uj_steps(self):
        self.service.modeling_obj_containers = ["uj_step"]
        with self.assertRaises(PermissionError):
            self.service.self_delete()

    def test_self_delete_removes_backward_links_and_recomputes_server_and_storage(self):
        self.server.modeling_obj_containers = [self.service]
        self.storage.modeling_obj_containers = [self.service]
        self.service.self_delete()
        self.assertEqual([], self.server.modeling_obj_containers)
        self.assertEqual([], self.storage.modeling_obj_containers)
        self.server.compute_calculated_attributes.assert_called_once()
        self.storage.compute_calculated_attributes.assert_called_once()


if __name__ == '__main__':
    unittest.main()



from efootprint.constants.units import u
from efootprint.core.hardware.network import Network
from efootprint.core.hardware.device_population import DevicePopulation
from efootprint.core.hardware.servers.server_base_class import Server
from efootprint.core.hardware.storage import Storage
from efootprint.core.service import Service
from efootprint.core.usage.usage_pattern import UsagePattern
from efootprint.abstract_modeling_classes.explainable_objects import ExplainableQuantity
from efootprint.utils.tools import format_co2_amount, display_co2_amount

from typing import Dict, List, Set
import plotly.express as px
import plotly
import pandas as pd


class System:
    def __init__(self, name: str, usage_patterns: List[UsagePattern]):
        self.name = name
        usage_pattern_names = [usage_pattern.name for usage_pattern in usage_patterns]
        if len(usage_pattern_names) != len(set(usage_pattern_names)):
            raise ValueError("You can’t have 2 usage patterns with the same name within a System")
        self._usage_patterns = usage_patterns

        self.launch_computations()

    @property
    def usage_patterns(self):
        return self._usage_patterns

    @usage_patterns.setter
    def usage_patterns(self, new_usage_patterns):
        self._usage_patterns = new_usage_patterns

        self.launch_computations()

    @property
    def servers(self) -> Set[Server]:
        output_set = set()
        for usage_pattern in self.usage_patterns:
            output_set.update(usage_pattern.user_journey.servers)

        return output_set

    @property
    def storages(self) -> Set[Storage]:
        output_set = set()
        for usage_pattern in self.usage_patterns:
            output_set.update(usage_pattern.user_journey.storages)

        return output_set

    @property
    def services(self) -> Set[Service]:
        output_set = set()
        for usage_pattern in self.usage_patterns:
            output_set.update(usage_pattern.user_journey.services)

        return output_set

    @property
    def device_populations(self) -> Set[DevicePopulation]:
        output_set = set()
        for usage_pattern in self.usage_patterns:
            output_set.update({usage_pattern.device_population})

        return output_set

    @property
    def networks(self) -> Set[Network]:
        output_set = set()
        for usage_pattern in self.usage_patterns:
            output_set.update({usage_pattern.network})

        return output_set

    def launch_computations(self):
        for usage_pattern in self.usage_patterns:
            usage_pattern.compute_calculated_attributes()
        for device_population in self.device_populations:
            device_population.compute_calculated_attributes()
        for service in self.services:
            service.compute_calculated_attributes()
        for network in self.networks:
            network.compute_calculated_attributes()
        for server in self.servers:
            server.compute_calculated_attributes()
        for storage in self.storages:
            storage.compute_calculated_attributes()

    def get_storage_by_name(self, storage_name) -> Storage:
        for storage in self.storages:
            if storage.name == storage_name:
                return storage

    def get_server_by_name(self, server_name) -> Server:
        for server in self.servers:
            if server.name == server_name:
                return server

    def get_usage_pattern_by_name(self, usage_pattern_name) -> UsagePattern:
        for usage_pattern in self.usage_patterns:
            if usage_pattern.name == usage_pattern_name:
                return usage_pattern

    def fabrication_footprints(self) -> Dict[str, Dict[str, ExplainableQuantity]]:
        fab_footprints = {
            "Servers": {server.name: server.instances_fabrication_footprint for server in self.servers},
            "Storage": {storage.name: storage.instances_fabrication_footprint for storage in self.storages},
            "Devices": {device_population.name: device_population.instances_fabrication_footprint
                        for device_population in self.device_populations},
            "Network": {"networks": ExplainableQuantity(0 * u.kg / u.year, "No fabrication footprint for networks")}
        }

        return fab_footprints

    def energy_footprints(self) -> Dict[str, Dict[str, ExplainableQuantity]]:
        energy_footprints = {
            "Servers": {server.name: server.energy_footprint for server in self.servers},
            "Storage": {storage.name: storage.energy_footprint for storage in self.storages},
            "Devices": {device_population.name: device_population.energy_footprint
                        for device_population in self.device_populations},
            "Network": {network.name: network.energy_footprint for network in self.networks},
        }

        return energy_footprints

    def total_fabrication_footprints(self) -> Dict[str, ExplainableQuantity]:
        fab_footprints = {
            "Servers": sum(server.instances_fabrication_footprint for server in self.servers),
            "Storage": sum(storage.instances_fabrication_footprint for storage in self.storages),
            "Devices": sum(device_population.instances_fabrication_footprint
                           for device_population in self.device_populations),
            "Network": ExplainableQuantity(0 * u.kg / u.year, "No fabrication footprint for networks")
        }

        return fab_footprints

    def total_energy_footprints(self) -> Dict[str, ExplainableQuantity]:
        energy_footprints = {
            "Servers": sum(server.energy_footprint for server in self.servers),
            "Storage": sum(storage.energy_footprint for storage in self.storages),
            "Devices": sum(device_population.energy_footprint for device_population in self.device_populations),
            "Network": sum(network.energy_footprint for network in self.networks)
        }

        return energy_footprints

    def total_footprint(self) -> ExplainableQuantity:
        return (
            sum(
                sum(
                    self.fabrication_footprints()[key].values()) + sum(self.energy_footprints()[key].values())
                for key in self.fabrication_footprints().keys()
            )
        ).set_label(f"{self.name} total carbon footprint")

    def plot_footprints_by_category_and_object(self, filename=None):
        fab_footprints = self.fabrication_footprints()
        energy_footprints = self.energy_footprints()
        categories = list(fab_footprints.keys())

        rows_as_dicts = []

        value_colname = "Carbon footprints in tons CO2eq / year"
        for category in categories:
            fab_objects = sorted(fab_footprints[category].items(), key=lambda x: x[0])
            energy_objects = sorted(energy_footprints[category].items(), key=lambda x: x[0])

            for objs, color in zip([energy_objects, fab_objects], ["Electricity", "Fabrication"]):
                data_dicts = [
                    {"Type": color, "Category": category, "Object": obj[0],
                     value_colname: obj[1].value.magnitude / 1000,
                     "Amount": f"{display_co2_amount(format_co2_amount(obj[1].value.magnitude))} / year"}
                    for obj in objs]
                rows_as_dicts += data_dicts

        df = pd.DataFrame.from_records(rows_as_dicts)

        total_co2 = df[value_colname].sum()

        fig = px.bar(
            df, x="Category", y=value_colname, color='Type', barmode='group', height=400,
            hover_data={"Type": False, "Category": False, "Object": True, value_colname: False, "Amount": True},
            template="plotly_white", width=800,
            title=f"Total CO2 emissions from {self.name}: {display_co2_amount(format_co2_amount(total_co2 * 1000))} / year")

        total_co2_per_category_and_type = df.groupby(["Category", "Type"])[value_colname].sum()

        for category, source_type in total_co2_per_category_and_type.keys():
            height = total_co2_per_category_and_type.loc[category, source_type]
            x_shift_direction = 1 if source_type == 'Fabrication' else -1

            fig.add_annotation(
                x=category,
                y=height,
                text=f"{int((height / total_co2) * 100)}%",  # Format the label as a percentage
                showarrow=False,
                yshift=10,  # Shift the label slightly above the stack
                xshift=30 * x_shift_direction
            )

        if filename is None:
            filename = f"{self.name} footprints"

        plotly.offline.plot(fig, filename=filename)



from efootprint.constants.units import u
from efootprint.core.hardware.device_population import DevicePopulation
from efootprint.core.usage.user_journey import UserJourney
from efootprint.core.service import Service
from efootprint.core.hardware.network import Network
from efootprint.abstract_modeling_classes.source_objects import SourceValue, SourceObject
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.abstract_modeling_classes.explainable_objects import ExplainableHourlyUsage

from typing import List


class UsagePattern(ModelingObject):
    @staticmethod
    def check_time_intervals_validity(sorted_time_intervals):
        for i in range(len(sorted_time_intervals)):
            start_time, end_time = sorted_time_intervals[i]
            if start_time >= end_time:
                raise ValueError(
                    f"Invalid time interval {sorted_time_intervals[i]}, start time must be earlier than end time")
            if i < len(sorted_time_intervals) - 1:
                next_start_time = sorted_time_intervals[i + 1][0]
                if next_start_time < end_time:
                    raise ValueError(
                        f"Time interval {sorted_time_intervals[i + 1]} starts before time interval"
                        f" {sorted_time_intervals[i]} ends")

    def __init__(self, name: str, user_journey: UserJourney, device_population: DevicePopulation,
                 network: Network, user_journey_freq_per_user: SourceValue, time_intervals: SourceObject):
        super().__init__(name)
        self.hourly_usage = None
        self.usage_time_fraction = None
        self.user_journey = user_journey
        self.device_population = device_population
        self.network = network
        if not user_journey_freq_per_user.value.check("[user_journey] / ([person] * [time])"):
            raise ValueError(f"User journey frequency defined in {self.name} should have "
                             f"[user_journey] / ([user] * [time]) dimensionality")
        self.user_journey_freq_per_user = user_journey_freq_per_user.set_label(f"Usage frequency in {self.name}")
        self.check_time_intervals_validity(time_intervals.value)
        self.time_intervals = time_intervals.set_label(f"{self.name} time intervals in local timezone")

        self.calculated_attributes = [
            "hourly_usage", "usage_time_fraction"]

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List[ModelingObject]:
        return [self.device_population]

    @property
    def services(self) -> List[Service]:
        return self.user_journey.services

    def update_hourly_usage(self):
        hourly_usage = [0 * u.dimensionless] * 24
        for time_interval in self.time_intervals.value:
            start, end = time_interval
            for i in range(start, end):
                hourly_usage[i] = 1 * u.dimensionless

        self.hourly_usage = ExplainableHourlyUsage(
            hourly_usage, f"{self.name} local timezone hourly usage", left_child=self.time_intervals,
            child_operator="Hourly usage conversion")

    def update_usage_time_fraction(self):
        usage_time_fraction = self.hourly_usage.compute_usage_time_fraction()
        self.usage_time_fraction = usage_time_fraction.set_label(
            f"Usage time fraction of {self.name}")

    @property
    def user_journey_freq(self):
        return self.device_population.user_journey_freq_per_up[self]

    @property
    def nb_user_journeys_in_parallel_during_usage(self):
        return self.device_population.nb_user_journeys_in_parallel_during_usage_per_up[self]

    @property
    def utc_time_intervals(self):
        return self.device_population.utc_time_intervals_per_up[self]



from efootprint.constants.units import u
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.core.service import Service
from efootprint.core.hardware.servers.server_base_class import Server
from efootprint.core.hardware.storage import Storage
from efootprint.logger import logger

from typing import List, Set, Type, Optional


class UserJourneyStep(ModelingObject):
    def __init__(self, name: str, service: Optional[Service], data_upload: SourceValue, data_download: SourceValue,
                 user_time_spent: SourceValue, request_duration: SourceValue = None,
                 cpu_needed: SourceValue = None, ram_needed: SourceValue = None):
        super().__init__(name)
        self.ram_needed = None
        self.service = service
        if not data_upload.value.check("[]/[user_journey]"):
            raise ValueError("Variable 'data_upload' does not have the appropriate '[]/[user_journey]' dimensionality")
        if not data_download.value.check("[]/[user_journey]"):
            raise ValueError("Variable 'data_upload' does not have the appropriate '[]/[user_journey]' dimensionality")
        self.data_upload = data_upload.set_label(f"Data upload of request {self.name}")
        self.data_download = data_download.set_label(f"Data download of request {self.name}")

        if not user_time_spent.value.check("[time]/[user_journey]"):
            raise ValueError(
                "Variable 'user_time_spent' does not have the appropriate '[time]/[user_journey]' dimensionality")
        self.user_time_spent = user_time_spent.set_label(f"Time spent on step {self.name}")

        if self.service is None:
            if (self.data_download.magnitude > 0 or self.data_upload.magnitude > 0 or request_duration
                    or ram_needed or cpu_needed):
                raise ValueError(
                    f"When creating a user journey without service there should be no data transfer and no request"
                    f" duration or resource need")
        else:
            if request_duration is None:
                request_duration = SourceValue(1 * u.s)
            if not request_duration.value.check("[time]"):
                raise ValueError("Variable 'request_duration' does not have the appropriate '[time]' dimensionality")
            self.request_duration = request_duration.set_label(f"Request duration to {self.service.name} in {self.name}")
            if request_duration.value > user_time_spent.value * u.uj:
                # TODO: define a setter method to make this check also when the attribute is updated
                logger.warning("Variable 'request_duration' is greater than variable 'user_time_spent'")

            if ram_needed is None:
                ram_needed = SourceValue(100 * u.MB / u.uj)
            if not ram_needed.value.check("[] / [user_journey]"):
                raise ValueError(
                    "Variable 'ram_needed' does not have the appropriate '[] / [user_journey]' dimensionality")
            self.ram_needed = ram_needed.set_label(f"RAM needed on server {self.service.server.name} to process {self.name}")

            if cpu_needed is None:
                cpu_needed = SourceValue(1 * u.core / u.uj)
            if not cpu_needed.value.check("[cpu] / [user_journey]"):
                raise ValueError(
                    "Variable 'cpu_needed' does not have the appropriate '[cpu] / [user_journey]' dimensionality")
            self.cpu_needed = cpu_needed.set_label(f"CPU needed on server {self.service.server.name} to process {self.name}")

    @property
    def user_journeys(self):
        return self.modeling_obj_containers

    @property
    def usage_patterns(self):
        return list(set(sum([uj.usage_patterns for uj in self.user_journeys], start=[])))

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List:
        if len(self.user_journeys) > 0:
            return self.user_journeys
        elif self.service is not None:
            return [self.service]
        else:
            return []

    def self_delete(self):
        logger.warning(f"Deleting {self.name} and removing backward link in its service")
        if self.user_journeys:
            raise PermissionError(f"You can’t delete {self.name} because it has a user_journey pointing to it")
        self.service.modeling_obj_containers = [elt for elt in self.service.modeling_obj_containers if elt != self]
        self.service.launch_attributes_computation_chain()

    def after_init(self):
        self.init_has_passed = True
        self.compute_calculated_attributes()

    def __mul__(self, other):
        if type(other) not in [int, float]:
            raise ValueError(f"Can only multiply UserJourneyStep with int or float, not {type(other)}")
        raise NotImplementedError


class UserJourney(ModelingObject):
    def __init__(self, name: str, uj_steps: List[UserJourneyStep]):
        super().__init__(name)
        self.data_upload = None
        self.data_download = None
        self.duration = None
        self.calculated_attributes = ["duration", "data_download", "data_upload"]
        self.uj_steps = uj_steps

    @property
    def servers(self) -> Set[Server]:
        servers = set()
        for uj_step in self.uj_steps:
            if uj_step.service is not None:
                servers = servers | {uj_step.service.server}

        return servers

    @property
    def storages(self) -> Set[Storage]:
        storages = set()
        for uj_step in self.uj_steps:
            if uj_step.service is not None:
                storages = storages | {uj_step.service.storage}

        return storages

    @property
    def services(self) -> List[Service]:
        services = set()
        for uj_step in self.uj_steps:
            if uj_step.service is not None:
                services = services | {uj_step.service}

        return list(services)

    @property
    def usage_patterns(self):
        return self.modeling_obj_containers

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List[Type["UsagePattern"]]:
        return self.usage_patterns

    def after_init(self):
        self.init_has_passed = True
        self.compute_calculated_attributes()

    def add_step(self, step: UserJourneyStep) -> None:
        step.add_obj_to_modeling_obj_containers(self)
        self.uj_steps = self.uj_steps + [step]

    def update_duration(self):
        user_time_spent_sum = sum([uj_step.user_time_spent for uj_step in self.uj_steps])

        self.duration = user_time_spent_sum.set_label(f"Duration of {self.name}")

    def update_data_download(self):
        all_data_download = 0
        for uj_step in self.uj_steps:
            all_data_download += uj_step.data_download

        self.data_download = all_data_download.set_label(f"Data download of {self.name}")

    def update_data_upload(self):
        all_data_upload = 0
        for uj_step in self.uj_steps:
            all_data_upload += uj_step.data_upload

        self.data_upload = all_data_upload.set_label(f"Data upload of {self.name}")



from efootprint.abstract_modeling_classes.explainable_objects import ExplainableQuantity
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.hardware.servers.server_base_class import Server

import math


class OnPremise(Server):
    def __init__(self, name: str, carbon_footprint_fabrication: SourceValue, power: SourceValue,
                 lifespan: SourceValue, idle_power: SourceValue, ram: SourceValue, cpu_cores: SourceValue,
                 power_usage_effectiveness: SourceValue, average_carbon_intensity: SourceValue,
                 server_utilization_rate: SourceValue):
        super().__init__(
            name, carbon_footprint_fabrication, power, lifespan, idle_power, ram, cpu_cores, power_usage_effectiveness,
            average_carbon_intensity, server_utilization_rate)

    def update_nb_of_instances(self):
        ram_needed_per_day = self.all_services_ram_needs.max().set_label(
            f"Max daily {self.name} RAM need")
        cpu_needed_per_day = self.all_services_cpu_needs.max().set_label(
            f"Max daily {self.name} CPU need")

        nb_of_servers_based_on_ram_alone = (
                ram_needed_per_day / self.available_ram_per_instance).set_label(
            f"Raw nb of {self.name} instances based on RAM alone")
        nb_of_servers_based_on_cpu_alone = (
                cpu_needed_per_day / self.available_cpu_per_instance).set_label(
            f"Raw nb of {self.name} instances based on CPU alone")

        nb_of_servers_raw = nb_of_servers_based_on_ram_alone.compare_with_and_return_max(
            nb_of_servers_based_on_cpu_alone)

        if math.ceil(nb_of_servers_raw.magnitude) - nb_of_servers_raw.magnitude != 0:
            nb_of_instances = nb_of_servers_raw + ExplainableQuantity(
                (math.ceil(nb_of_servers_raw.magnitude) - nb_of_servers_raw.magnitude)
                * u.dimensionless, "Extra server capacity because number of servers must be an integer")
        else:
            nb_of_instances = nb_of_servers_raw

        self.nb_of_instances = nb_of_instances.set_label(f"Nb of {self.name} instances")

    def update_instances_power(self):
        effective_active_power = self.power * self.power_usage_effectiveness
        effective_idle_power = self.idle_power * self.power_usage_effectiveness

        fraction_of_time_not_in_use = ExplainableQuantity(1 * u.dimensionless, "100%") - self.fraction_of_time_in_use
        server_power = (
                self.nb_of_instances *
                ((effective_active_power * self.fraction_of_time_in_use)
                 + (effective_idle_power * fraction_of_time_not_in_use))
        ).to(u.kWh / u.year)

        self.instances_power = server_power.set_label(f"Power of {self.name} instances")



from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.core.usage.user_journey import UserJourneyStep, UserJourney
from efootprint.constants.units import u

import unittest
from unittest import TestCase
from unittest.mock import MagicMock, patch, PropertyMock


class TestUserJourneyStep(TestCase):
    def setUp(self):
        self.service = MagicMock()

        self.user_journey_step = UserJourneyStep(
            "test uj step", service=self.service, data_download=SourceValue(200 * u.MB / u.uj),
            data_upload=SourceValue(100 * u.MB / u.uj),
            ram_needed=SourceValue(400 * u.MB / u.uj), cpu_needed=SourceValue(2 * u.core / u.uj),
            user_time_spent=SourceValue(2 * u.min / u.uj))

    def test_user_journey_step_with_null_service_doesnt_break(self):
        uj_step_without_service = UserJourneyStep(
            "", service=None, data_download=SourceValue(0 * u.MB / u.uj),
            data_upload=SourceValue(0 * u.MB / u.uj),
            user_time_spent=SourceValue(2 * u.min / u.uj))

    def test_user_journey_step_with_null_service_breaks_if_non_null_data_transfer(self):
        with self.assertRaises(ValueError):
            uj_step_without_service = UserJourneyStep(
                "", service=None, data_download=SourceValue(10 * u.MB / u.uj),
                data_upload=SourceValue(0 * u.MB / u.uj),
                user_time_spent=SourceValue(2 * u.min / u.uj))

    def test_self_delete_should_raise_error_if_self_has_associated_uj(self):
        self.user_journey_step.modeling_obj_containers = ["uj"]
        with self.assertRaises(PermissionError):
            self.user_journey_step.self_delete()

    def test_self_delete_removes_backward_links_and_recomputes_server_and_storage(self):
        self.service.modeling_obj_containers = [self.user_journey_step]
        self.user_journey_step.self_delete()
        self.assertEqual([], self.service.modeling_obj_containers)
        self.service.launch_attributes_computation_chain.assert_called_once()


class TestUserJourney(TestCase):
    def setUp(self):
        self.storage = MagicMock()
        self.server = MagicMock()
        self.service = MagicMock()
        self.service.server = self.server
        self.service.storage = self.storage

        self.step1 = UserJourneyStep(
            "test_uj_step1", service=self.service, data_download=SourceValue(200 * u.MB / u.uj),
            data_upload=SourceValue(100 * u.MB / u.uj),
            ram_needed=SourceValue(400 * u.MB / u.uj), cpu_needed=SourceValue(2 * u.core / u.uj),
            user_time_spent=SourceValue(2 * u.min / u.uj))
        self.step2 = UserJourneyStep(
            "test_uj_step2", service=self.service, data_download=SourceValue(200 * u.MB / u.uj),
            data_upload=SourceValue(100 * u.MB / u.uj),
            ram_needed=SourceValue(400 * u.MB / u.uj), cpu_needed=SourceValue(2 * u.core / u.uj),
            user_time_spent=SourceValue(2 * u.min / u.uj))
        self.step3 = UserJourneyStep(
            "test_uj_step3", service=self.service, data_download=SourceValue(200 * u.MB / u.uj),
            data_upload=SourceValue(100 * u.MB / u.uj),
            ram_needed=SourceValue(400 * u.MB / u.uj), cpu_needed=SourceValue(2 * u.core / u.uj),
            user_time_spent=SourceValue(2 * u.min / u.uj))
        self.one_user_journey = SourceValue(1 * u.user_journey)
        self.user_journey = UserJourney("test user journey", uj_steps=[self.step1])
        self.user_journey.dont_handle_input_updates = True
        self.usage_pattern = MagicMock()
        self.user_journey.modeling_obj_containers = [self.usage_pattern]

        for step in [self.step1, self.step2, self.step3]:
            step.remove_obj_from_modeling_obj_containers = MagicMock()
            step.add_obj_to_modeling_obj_containers = MagicMock()
            step.launch_attributes_computation_chain = MagicMock()

    def test_servers(self):
        self.assertEqual(self.user_journey.servers, {self.server})

    def test_storages(self):
        self.assertEqual(self.user_journey.storages, {self.storage})

    def test_services(self):
        self.assertEqual(self.user_journey.services, [self.service])

    def test_update_duration_with_multiple_steps(self):
        self.user_journey.add_step(self.step1)
        for step in self.user_journey.uj_steps:
            step.user_time_spent = SourceValue(5 * u.min / u.user_journey)

        self.user_journey.update_duration()
        expected_duration = SourceValue(10 * u.min / u.user_journey)

        self.assertEqual(self.user_journey.duration.value, expected_duration.value)

    def test_update_data_download_with_multiple_steps(self):
        self.user_journey.add_step(self.step1)

        for step in self.user_journey.uj_steps:
            step.data_download = SourceValue(10 * u.MB / u.user_journey)
        self.user_journey.update_data_download()

        expected_data_download = SourceValue(20 * u.MB / u.user_journey)

        self.assertEqual(self.user_journey.data_download.value, expected_data_download.value)

    def test_update_data_upload_with_multiple_steps(self):
        self.user_journey.add_step(self.step1)
        for step in self.user_journey.uj_steps:
            step.data_upload = SourceValue(10 * u.MB / u.user_journey)

        self.user_journey.update_data_upload()

        expected_data_upload = SourceValue(20 * u.MB / u.user_journey)

        self.assertEqual(self.user_journey.data_upload.value, expected_data_upload.value)


if __name__ == "__main__":
    unittest.main()



from setuptools import setup, find_packages
import os
import codecs


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

setup(
    name="efootprint",
    version=get_version(os.path.join("efootprint", "__init__.py")),
    author="Vincent Villet for Publicis Sapient",
    author_email="vincent.villet@gmail.com",
    description="Digital service environmental footprint model",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    url="https://github.com/publicissapient-france/e-footprint",
    packages=find_packages(exclude="tests"),
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
    # TODO: Remove matplotlib from requirements to save bandwidth as usage grows ?
    install_requires=read('requirements.txt').split('\n'),
    include_package_data=True
)



from efootprint.constants.countries import Countries
from efootprint.core.hardware.hardware_base_classes import InfraHardware
from efootprint.abstract_modeling_classes.explainable_objects import ExplainableQuantity
from efootprint.constants.sources import Sources
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u


class Storage(InfraHardware):
    def __init__(self, name: str, carbon_footprint_fabrication: SourceValue, power: SourceValue,
                 lifespan: SourceValue, idle_power: SourceValue, storage_capacity: SourceValue,
                 power_usage_effectiveness: SourceValue, average_carbon_intensity: SourceValue,
                 data_replication_factor: SourceValue, storage_need_from_previous_year: SourceValue = None):
        super().__init__(name, carbon_footprint_fabrication, power, lifespan, average_carbon_intensity)
        self.all_services_storage_needs = None
        self.long_term_storage_required = None
        self.active_storage_required = None
        self.nb_of_idle_instances = None
        self.nb_of_active_instances = None
        self.instances_power = None
        self.idle_power = idle_power.set_label(f"Idle power of {self.name}")
        self.storage_capacity = storage_capacity.set_label(f"Storage capacity of {self.name}")
        self.power_usage_effectiveness = power_usage_effectiveness.set_label(f"PUE of {self.name}")
        self.data_replication_factor = data_replication_factor.set_label(f"Data replication factor of {self.name}")
        # TODO: implement data storage duration logic
        if storage_need_from_previous_year is not None:
            if not storage_need_from_previous_year.value.check("[]"):
                raise ValueError(
                    "Value of variable 'storage_need_from_previous_year' does not have the appropriate"
                    " '[]' dimensionality")
            storage_need_from_previous_year.left_parent = None
            storage_need_from_previous_year.right_parent = None
            storage_need_from_previous_year.label = f"{self.name} storage need from previous year"
            self.storage_need_from_previous_year = storage_need_from_previous_year
        else:
            self.storage_need_from_previous_year = 0

        self.calculated_attributes = [
            "all_services_storage_needs", "active_storage_required", "long_term_storage_required",
            "nb_of_idle_instances", "nb_of_active_instances"
        ] + self.calculated_attributes_defined_in_infra_hardware_class

    def update_all_services_storage_needs(self):
        if len(self.services) > 0:
            all_services_storage_needs = sum(service.storage_needed for service in self.services).to(u.TB / u.year)

            self.all_services_storage_needs = all_services_storage_needs.set_label(
                f"Storage need of {self.name}")
        else:
            self.all_services_storage_needs = ExplainableQuantity(
                0 * u.TB / u.year, f"No storage need for {self.name} because no associated services")

    def update_long_term_storage_required(self):
        # TODO: Higher level year by year analysis that adds storage from previous year with remaining storage duration
        # to current storage
        long_term_storage_required = (
                self.all_services_storage_needs * self.data_replication_factor
                * ExplainableQuantity(1 * u.year, "one year") + self.storage_need_from_previous_year
                - self.active_storage_required
        )

        self.long_term_storage_required = long_term_storage_required.set_label(
            f"Long term storage required for {self.name}")

    def update_active_storage_required(self):
        active_storage_required = (
                self.all_services_storage_needs
                * SourceValue(1 * u.hour, Sources.HYPOTHESIS, "Time interval during which active storage is considered")
        ).to(u.GB)

        self.active_storage_required = active_storage_required.set_label(
            f"Active storage required for {self.name}")

    def update_nb_of_idle_instances(self):
        nb_of_idle_instances = (self.long_term_storage_required / self.storage_capacity).to(u.dimensionless)

        self.nb_of_idle_instances = nb_of_idle_instances.set_label(
            f"Number of idle storage units for {self.name}")

    def update_nb_of_active_instances(self):
        active_instances = (self.active_storage_required / self.storage_capacity).to(u.dimensionless)

        self.nb_of_active_instances = active_instances.set_label(
            f"Number of active instances for {self.name}")

    def update_nb_of_instances(self):
        nb_of_instances = self.nb_of_active_instances + self.nb_of_idle_instances

        self.nb_of_instances = nb_of_instances.set_label(
            f"Number of total instances for {self.name}")

    def update_instances_power(self):
        active_storage_power = (
                self.nb_of_active_instances * self.power * self.power_usage_effectiveness
        ).set_label(f"Active instances power")
        idle_storage_power = (
                self.nb_of_idle_instances * self.idle_power * self.power_usage_effectiveness
        ).set_label(f"Idle instances power")
        storage_power = (
                active_storage_power * self.fraction_of_time_in_use + idle_storage_power).to(u.kWh / u.year)

        self.instances_power = storage_power.set_label(
            f"Storage power for {self.name}")



from efootprint.constants.sources import Sources
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.hardware.servers.autoscaling import Autoscaling
from tests.utils import create_cpu_need, create_ram_need

from unittest import TestCase
from unittest.mock import MagicMock, patch


class TestAutoscaling(TestCase):
    def setUp(self):
        self.country = MagicMock()
        self.server_base = Autoscaling(
            "Autoscaling",
            carbon_footprint_fabrication=SourceValue(0 * u.kg, Sources.BASE_ADEME_V19),
            power=SourceValue(0 * u.W, Sources.HYPOTHESIS),
            lifespan=SourceValue(0 * u.year, Sources.HYPOTHESIS),
            idle_power=SourceValue(0 * u.W, Sources.HYPOTHESIS),
            ram=SourceValue(0 * u.GB, Sources.HYPOTHESIS),
            cpu_cores=SourceValue(0 * u.core, Sources.HYPOTHESIS),
            power_usage_effectiveness=SourceValue(0 * u.dimensionless, Sources.HYPOTHESIS),
            average_carbon_intensity=SourceValue(100 * u.g / u.kWh),
            server_utilization_rate=SourceValue(0 * u.dimensionless)
        )
        self.server_base.dont_handle_input_updates = True

    def test_nb_of_instances_autoscaling(self):
        ram_need = (create_ram_need([[0, 12]], 100 * u.GB) + create_ram_need([[12, 24]], 150 * u.GB))
        cpu_need = create_cpu_need([[0, 24]], 1 * u.core)

        with patch.object(self.server_base, "all_services_ram_needs", new=ram_need), \
                patch.object(self.server_base, "all_services_cpu_needs", new=cpu_need), \
                patch.object(self.server_base, "available_ram_per_instance", new=SourceValue(100 * u.GB)), \
                patch.object(self.server_base, "available_cpu_per_instance", new=SourceValue(25 * u.core)):
            self.server_base.update_nb_of_instances()
            self.assertEqual(1.5 * u.dimensionless, round(self.server_base.nb_of_instances.value, 2))

    def test_compute_instances_power_cloud_autoscaling(self):
        with patch.object(self.server_base,
                          "fraction_of_time_in_use", SourceValue(((24 - 10) / 24) * u.dimensionless)), \
                patch.object(self.server_base, "nb_of_instances", SourceValue(10 * u.dimensionless)), \
                patch.object(self.server_base, "power", SourceValue(300 * u.W)), \
                patch.object(self.server_base, "idle_power", SourceValue(50 * u.W)), \
                patch.object(self.server_base, "power_usage_effectiveness", SourceValue(1.2 * u.dimensionless)):
            self.server_base.update_instances_power()
            self.assertEqual(round((3600 * u.W).to(u.kWh / u.year), 2),
                             round(self.server_base.instances_power.value, 2))



WIDTH = 1200
HEIGHT = 900


def set_string_max_width(s, max_width):
    lines = s.split('\n')
    formatted_lines = []

    for line in lines:
        words = line.split()
        current_line = []
        current_len = 0

        for word in words:
            if current_len + len(word) > max_width:
                formatted_lines.append(' '.join(current_line))
                current_line = [word]
                current_len = len(word) + 1
            else:
                current_line.append(word)
                current_len += len(word) + 1

        if current_line:
            formatted_lines.append(' '.join(current_line))

    return '\n'.join(formatted_lines)



from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.utils.graph_tools import WIDTH, HEIGHT, set_string_max_width
from efootprint.utils.tools import convert_to_list

from pyvis.network import Network

COLOR_MAP = {
    "Autoscaling": "red",
    "OnPremise": "red",
    "Serverless": "red",
    "Hardware": "red",
    "Storage": "red",
    "Service": "deepskyblue",
    "UsagePattern": "blueviolet",
    "DevicePopulation": "red",
    "UserJourney": "gold",
    "UserJourneyStep": "gold",
    "TimeIntervals": "gold",
}

USAGE_PATTERN_VIEW_CLASSES_TO_IGNORE = [
    "System", "TimeIntervals", "Network", "Hardware", "Country"]
SERVICES_VIEW_CLASSES_TO_IGNORE = [
    "System", "UsagePattern", "TimeIntervals", "Network", "Server", "Storage", "Hardware", "DevicePopulation"]
SERVICES_AND_INFRA_VIEW_CLASSES_TO_IGNORE = [
    "UsagePattern", "TimeIntervals", "Network", "Hardware", "DevicePopulation", "System",
    "UserJourney", "UserJourneyStep"]


def build_object_relationships_graph(
        node, input_graph=None, visited=None, classes_to_ignore=None, width=WIDTH, height=HEIGHT):
    if classes_to_ignore is None:
        classes_to_ignore = ["System"]
    if input_graph is None:
        input_graph = Network(notebook=True, width=f"{width}px", height=f"{height}px")
    if visited is None:
        visited = set()

    if node in visited:
        return input_graph

    node_type = type(node).__name__
    if node_type not in classes_to_ignore:
        input_graph.add_node(
            id(node), label=set_string_max_width(f"{node.name}", 20),
            color=COLOR_MAP.get(node_type, "gray"))

    for attr_name, attr_value in vars(node).items():
        values = convert_to_list(attr_value)
        for value in values:
            if isinstance(value, ModelingObject):
                value_type = type(value).__name__
                if value_type not in classes_to_ignore:
                    input_graph.add_node(
                        id(value), label=set_string_max_width(f"{value.name}", 20),
                        color=COLOR_MAP.get(value_type, "gray"))
                    if node_type not in classes_to_ignore:
                        input_graph.add_edge(id(node), id(value), title=attr_name)

                if value not in visited:
                    visited.add(node)
                    build_object_relationships_graph(value, input_graph, visited, classes_to_ignore, width, height)

    return input_graph



from efootprint.core.hardware.hardware_base_classes import InfraHardware
from efootprint.abstract_modeling_classes.explainable_objects import ExplainableHourlyUsage
from efootprint.constants.sources import Sources
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from tests.utils import create_cpu_need, create_ram_need

from copy import deepcopy
from unittest import TestCase
from unittest.mock import MagicMock, patch


class TestInfraHardware(TestCase):
    def setUp(self):
        class InfraHardwareTestClass(InfraHardware):
            def __init__(self, name: str, carbon_footprint_fabrication: SourceValue, power: SourceValue,
                         lifespan: SourceValue, average_carbon_intensity: SourceValue):
                super().__init__(name, carbon_footprint_fabrication, power, lifespan, average_carbon_intensity)
                self.dont_handle_pubsub_topic_messages = True

            def update_nb_of_instances(self):
                self.nb_of_instances = SourceValue(2 * u.dimensionless)

            def update_instances_power(self):
                self.instances_power = SourceValue(16 * u.W)

        self.test_infra_hardware = InfraHardwareTestClass(
            "test_infra_hardware", carbon_footprint_fabrication=SourceValue(120 * u.kg, Sources.USER_DATA),
            power=SourceValue(2 * u.W, Sources.USER_DATA), lifespan=SourceValue(6 * u.years, Sources.HYPOTHESIS),
            average_carbon_intensity=SourceValue(100 * u.g / u.kWh))

        self.service1 = MagicMock()
        self.ram_needs_service1 = create_ram_need([[0, 8]])
        self.cpu_needs_service1 = create_cpu_need([[0, 8]])
        self.test_infra_hardware_single_service = deepcopy(self.test_infra_hardware)
        self.service1.hour_by_hour_ram_need = self.ram_needs_service1
        self.service1.hour_by_hour_cpu_need = self.cpu_needs_service1
        self.test_infra_hardware_single_service.modeling_obj_containers = [self.service1]

        self.service2 = MagicMock()
        self.service3 = MagicMock()
        self.ram_needs_service2 = create_ram_need([[6, 14]])
        self.ram_needs_service3 = create_ram_need([[8, 16]])
        self.cpu_needs_service2 = create_cpu_need([[6, 14]])
        self.cpu_needs_service3 = create_cpu_need([[8, 16]])
        self.test_infra_hardware_multiple_services = deepcopy(self.test_infra_hardware)
        self.service2.hour_by_hour_ram_need = self.ram_needs_service2
        self.service3.hour_by_hour_ram_need = self.ram_needs_service3
        self.service2.hour_by_hour_cpu_need = self.cpu_needs_service2
        self.service3.hour_by_hour_cpu_need = self.cpu_needs_service3
        self.test_infra_hardware_multiple_services.modeling_obj_containers = [self.service2, self.service3]

    def test_update_all_services_ram_needs_single_service(self):
        expected_value = ExplainableHourlyUsage([0 * u.GB] * 24, "expected hourly usage")
        for i in range(8):
            expected_value.value[i] = 100 * u.GB

        self.test_infra_hardware_single_service.update_all_services_ram_needs()
        self.assertEqual(expected_value.value, self.test_infra_hardware_single_service.all_services_ram_needs.value)

    def test_services(self):
        service1 = MagicMock()
        service2 = MagicMock()
        with patch.object(
                self.test_infra_hardware_multiple_services, "modeling_obj_containers", new=[service1, service2]):
            self.assertEqual([service1, service2], self.test_infra_hardware_multiple_services.services)

    def test_all_services_infra_needs_multiple_services(self):
        expected_value = ExplainableHourlyUsage([0 * u.GB] * 24, "expected hourly usage")
        for i in range(6, 8):
            expected_value.value[i] = 100 * u.GB
        for i in range(8, 14):
            expected_value.value[i] = 200 * u.GB
        for i in range(14, 16):
            expected_value.value[i] = 100 * u.GB

        self.test_infra_hardware_multiple_services.update_all_services_ram_needs()
        self.assertEqual(expected_value.value, self.test_infra_hardware_multiple_services.all_services_ram_needs.value)

    def test_all_services_cpu_needs_single_service(self):
        expected_value = ExplainableHourlyUsage([0 * u.core] * 24, "expected hourly usage")
        for i in range(8):
            expected_value.value[i] = 1 * u.core

        self.test_infra_hardware_single_service.update_all_services_cpu_needs()
        self.assertEqual(expected_value.value, self.test_infra_hardware_single_service.all_services_cpu_needs.value)

    def test_all_services_cpu_needs_multiple_services(self):
        expected_value = ExplainableHourlyUsage([0 * u.core] * 24, "expected hourly usage")
        for i in range(6, 8):
            expected_value.value[i] = 1 * u.core
        for i in range(8, 14):
            expected_value.value[i] = 2 * u.core
        for i in range(14, 16):
            expected_value.value[i] = 1 * u.core

        self.test_infra_hardware_multiple_services.update_all_services_cpu_needs()
        self.assertEqual(expected_value.value, self.test_infra_hardware_multiple_services.all_services_cpu_needs.value)

    def test_fraction_of_time_in_use(self):
        self.test_infra_hardware_multiple_services.all_services_cpu_needs = create_cpu_need([[10, 15]])
        self.test_infra_hardware_multiple_services.all_services_ram_needs = create_ram_need([[0, 12]])
        expected_value = SourceValue((15 / 24) * u.dimensionless, label="fraction_of_time_in_use")
        self.test_infra_hardware_multiple_services.update_fraction_of_time_in_use()
        self.assertEqual(expected_value.value, self.test_infra_hardware_multiple_services.fraction_of_time_in_use.value)

    def test_instances_fabrication_footprint(self):
        self.test_infra_hardware_single_service.update_nb_of_instances()
        self.test_infra_hardware_single_service.update_instances_fabrication_footprint()
        self.assertEqual(
            40 * u.kg / u.year, self.test_infra_hardware_single_service.instances_fabrication_footprint.value)

    def test_energy_footprints(self):
        self.test_infra_hardware_single_service.update_instances_power()
        self.test_infra_hardware_single_service.update_energy_footprint()
        self.assertEqual(
            round(16 * 24 * 365.25 * 100 * 1e-6 * u.kg / u.year, 2),
            round(self.test_infra_hardware_single_service.energy_footprint.value, 2))



from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject
from efootprint.constants.sources import Sources
from efootprint.abstract_modeling_classes.source_objects import SourceValue, SourceObject
from efootprint.core.usage.user_journey import UserJourney, UserJourneyStep
from efootprint.core.hardware.servers.autoscaling import Autoscaling
from efootprint.core.hardware.storage import Storage
from efootprint.core.service import Service
from efootprint.core.hardware.device_population import DevicePopulation
from efootprint.core.usage.usage_pattern import UsagePattern
from efootprint.core.hardware.network import Network
from efootprint.core.system import System
from efootprint.constants.countries import Countries
from efootprint.constants.units import u
from efootprint.abstract_modeling_classes.modeling_object import get_subclass_attributes, ModelingObject
from efootprint.logger import logger
from efootprint.utils.calculus_graph import build_calculus_graph
from efootprint.utils.object_relationships_graphs import build_object_relationships_graph, \
    USAGE_PATTERN_VIEW_CLASSES_TO_IGNORE
from efootprint.builders.hardware.devices_defaults import default_laptop, default_screen
from tests.integration_tests.integration_test_base_class import IntegrationTestBaseClass

from copy import deepcopy
from typing import List
import os


class IntegrationTest(IntegrationTestBaseClass):
    @classmethod
    def setUpClass(cls):
        cls.server = Autoscaling(
            "Default server",
            carbon_footprint_fabrication=SourceValue(600 * u.kg, Sources.BASE_ADEME_V19),
            power=SourceValue(300 * u.W, Sources.HYPOTHESIS),
            lifespan=SourceValue(6 * u.year, Sources.HYPOTHESIS),
            idle_power=SourceValue(50 * u.W, Sources.HYPOTHESIS),
            ram=SourceValue(128 * u.GB, Sources.USER_DATA),
            cpu_cores=SourceValue(24 * u.core, Sources.USER_DATA),
            power_usage_effectiveness=SourceValue(1.2 * u.dimensionless, Sources.USER_DATA),
            average_carbon_intensity=SourceValue(100 * u.g / u.kWh, Sources.USER_DATA),
            server_utilization_rate=SourceValue(0.9 * u.dimensionless, Sources.HYPOTHESIS)
        )
        cls.storage = Storage(
            "Default SSD storage",
            carbon_footprint_fabrication=SourceValue(160 * u.kg, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            power=SourceValue(1.3 * u.W, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            lifespan=SourceValue(6 * u.years, Sources.HYPOTHESIS),
            idle_power=SourceValue(0.1 * u.W, Sources.HYPOTHESIS),
            storage_capacity=SourceValue(1 * u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            power_usage_effectiveness=SourceValue(1.2 * u.dimensionless),
            average_carbon_intensity=SourceValue(100 * u.g / u.kWh, Sources.HYPOTHESIS),
            data_replication_factor=SourceValue(3 * u.dimensionless)
        )
        cls.service = Service(
            "Youtube", cls.server, cls.storage, base_ram_consumption=SourceValue(300 * u.MB, Sources.HYPOTHESIS),
            base_cpu_consumption=SourceValue(2 * u.core, Sources.HYPOTHESIS))

        cls.streaming_step = UserJourneyStep(
            "20 min streaming on Youtube", cls.service, SourceValue(50 * u.kB / u.uj),
            SourceValue((2.5 / 3) * u.GB / u.uj),
            user_time_spent=SourceValue(20 * u.min / u.uj), request_duration=SourceValue(4 * u.min))
        cls.upload_step = UserJourneyStep(
            "0.4s of upload", cls.service, SourceValue(300 * u.kB / u.uj), SourceValue(0 * u.kB / u.uj),
            user_time_spent=SourceValue(1 * u.s / u.uj), request_duration=SourceValue(0.1 * u.s))

        cls.uj = UserJourney("Daily Youtube usage", uj_steps=[cls.streaming_step, cls.upload_step])
        cls.device_population = DevicePopulation(
            "French Youtube users on laptop", SourceValue(4e7 * 0.3 * u.user), Countries.FRANCE,
            [default_laptop()])

        cls.network = Network("Default network", SourceValue(0.05 * u("kWh/GB"), Sources.TRAFICOM_STUDY))
        cls.usage_pattern = UsagePattern(
            "Youtube usage in France", cls.uj, cls.device_population,
            cls.network, SourceValue(365 * u.user_journey / (u.user * u.year)),
            SourceObject([[7, 23]], Sources.USER_DATA))

        cls.system = System("system 1", [cls.usage_pattern])

        cls.initial_footprint = cls.system.total_footprint()
        cls.initial_fab_footprints = {
            cls.storage: cls.storage.instances_fabrication_footprint,
            cls.server: cls.server.instances_fabrication_footprint,
            cls.device_population: cls.device_population.instances_fabrication_footprint,
        }
        cls.initial_energy_footprints = {
            cls.storage: cls.storage.energy_footprint,
            cls.server: cls.server.energy_footprint,
            cls.network: cls.network.energy_footprint,
            cls.device_population: cls.device_population.energy_footprint,
        }

    def test_calculation_graph(self):
        graph = build_calculus_graph(self.system.total_footprint())
        graph.show(os.path.join(os.path.abspath(os.path.dirname(__file__)), "full_calculation_graph.html"))

    def test_object_relationship_graph(self):
        object_relationships_graph = build_object_relationships_graph(
            self.system, classes_to_ignore=USAGE_PATTERN_VIEW_CLASSES_TO_IGNORE)
        object_relationships_graph.show(
            os.path.join(os.path.abspath(os.path.dirname(__file__)), "object_relationships_graph.html"))

    def footprint_has_changed(self, objects_to_test: List[ModelingObject]):
        for obj in objects_to_test:
            try:
                initial_energy_footprint = round(self.initial_energy_footprints[obj].value, 2)
                self.assertNotEqual(initial_energy_footprint, obj.energy_footprint.value)
                if type(obj) != Network:
                    initial_fab_footprint = round(self.initial_fab_footprints[obj].value, 2)
                    new_footprint = round(obj.instances_fabrication_footprint.value + obj.energy_footprint.value, 2)
                    self.assertNotEqual(
                        initial_fab_footprint + initial_energy_footprint, new_footprint)
                    logger.info(
                        f"{obj.name} footprint has changed from {initial_fab_footprint + initial_energy_footprint}"
                        f" to {new_footprint}")
                else:
                    logger.info(f"{obj.name} footprint has changed from "
                                f"{round(initial_energy_footprint, 2)} to {round(obj.energy_footprint.value, 2)}")
            except AssertionError:
                raise AssertionError(f"Footprint hasn’t changed for {obj.name}")

    def footprint_has_not_changed(self, objects_to_test: List[ModelingObject]):
        for obj in objects_to_test:
            try:
                initial_energy_footprint = self.initial_energy_footprints[obj].value
                if type(obj) != Network:
                    initial_fab_footprint = self.initial_fab_footprints[obj].value
                    self.assertEqual(initial_fab_footprint, obj.instances_fabrication_footprint.value)
                self.assertEqual(initial_energy_footprint, obj.energy_footprint.value)
                logger.info(f"{obj.name} footprint is the same as in setup")
            except AssertionError:
                raise AssertionError(f"Footprint has changed for {obj.name}")

    def test_variations_on_inputs(self):
        def test_variations_on_obj_inputs(input_object: ModelingObject, attrs_to_skip=None, special_mult=None):
            if attrs_to_skip is None:
                attrs_to_skip = []
            logger.warning(f"Testing input variations on {input_object.name}")
            for expl_attr_name, expl_attr in get_subclass_attributes(input_object, ExplainableObject).items():
                if expl_attr.left_parent is None and expl_attr.right_parent is None \
                        and expl_attr_name not in attrs_to_skip:
                    old_value = expl_attr.value
                    expl_attr_new_value = deepcopy(expl_attr)
                    if special_mult and expl_attr_name in special_mult.keys():
                        expl_attr_new_value.value *= special_mult[expl_attr_name] * u.dimensionless
                    else:
                        expl_attr_new_value.value *= 100 * u.dimensionless
                    expl_attr_new_value.label = expl_attr.label
                    logger.info(f"{expl_attr_new_value.label} changing from {round(old_value, 1)} to"
                                f" {round(expl_attr_new_value.value, 1)}")
                    input_object.__setattr__(expl_attr_name, expl_attr_new_value)
                    new_footprint = self.system.total_footprint()
                    logger.info(f"system footprint went from {round(self.initial_footprint.value, 1)} "
                                f"to {round(new_footprint.value, 1)}")
                    assert round(self.initial_footprint.magnitude, 2) != round(new_footprint.magnitude, 2)
                    input_object.__setattr__(expl_attr_name, expl_attr)
                    assert round(self.system.total_footprint().magnitude, 2) == \
                           round(self.initial_footprint.magnitude, 2)

        test_variations_on_obj_inputs(self.streaming_step)
        test_variations_on_obj_inputs(
            self.server, attrs_to_skip=["fraction_of_usage_time", "idle_power"],
            special_mult={"ram": 0.01})
        test_variations_on_obj_inputs(
            self.storage, attrs_to_skip=["fraction_of_usage_time"])
        test_variations_on_obj_inputs(
            self.service,
            special_mult={"base_ram_consumption": 380, "base_cpu_consumption": 10})
        test_variations_on_obj_inputs(self.uj)
        test_variations_on_obj_inputs(self.device_population)
        test_variations_on_obj_inputs(self.network)
        test_variations_on_obj_inputs(
            self.usage_pattern, attrs_to_skip=["time_intervals"])

    def test_time_intervals_change(self):
        logger.warning("Updating time intervals in usage pattern")
        old_time_intervals = deepcopy(self.usage_pattern.time_intervals)
        calculus_graph = build_calculus_graph(self.server.energy_footprint)
        calculus_graph.show(os.path.join(".", "debug_calculus_graph.html"))
        self.usage_pattern.time_intervals = SourceObject([[7, 13]], Sources.USER_DATA)
        assert round(self.initial_footprint.magnitude, 2) != round(self.system.total_footprint().magnitude, 2)
        logger.warning("Setting time intervals back to initial value in usage pattern")
        self.usage_pattern.time_intervals = old_time_intervals
        assert round(self.initial_footprint.magnitude, 2) == round(self.system.total_footprint().magnitude, 2)

    def test_uj_step_update(self):
        logger.warning("Updating uj steps in default user journey")
        self.uj.uj_steps = [self.streaming_step]
        assert round(self.initial_footprint.magnitude, 2) != round(self.system.total_footprint().magnitude, 2)
        self.uj.uj_steps = [self.streaming_step, self.upload_step]
        assert round(self.initial_footprint.magnitude, 2) == round(self.system.total_footprint().magnitude, 2)

    def test_device_pop_update(self):
        logger.warning("Updating devices in device population")
        self.device_population.devices = [default_laptop(), default_screen()]
        assert round(self.initial_footprint.magnitude, 2) != round(self.system.total_footprint().magnitude, 2)
        self.device_population.devices = [default_laptop()]
        assert round(self.initial_footprint.magnitude, 2) == round(self.system.total_footprint().magnitude, 2)

    def test_update_server(self):
        new_server = Autoscaling(
            "New server, identical in specs to default one",
            carbon_footprint_fabrication=SourceValue(600 * u.kg, Sources.BASE_ADEME_V19),
            power=SourceValue(300 * u.W, Sources.HYPOTHESIS),
            lifespan=SourceValue(6 * u.year, Sources.HYPOTHESIS),
            idle_power=SourceValue(50 * u.W, Sources.HYPOTHESIS),
            ram=SourceValue(128 * u.GB, Sources.HYPOTHESIS),
            cpu_cores=SourceValue(24 * u.core, Sources.HYPOTHESIS),
            power_usage_effectiveness=SourceValue(1.2 * u.dimensionless, Sources.HYPOTHESIS),
            average_carbon_intensity=SourceValue(100 * u.g / u.kWh, Sources.HYPOTHESIS),
            server_utilization_rate=SourceValue(0.9 * u.dimensionless, Sources.HYPOTHESIS)
        )

        logger.warning("Changing service server")
        self.service.server = new_server

        self.assertEqual(0, self.server.instances_fabrication_footprint.magnitude)
        self.assertEqual(0, self.server.energy_footprint.magnitude)
        self.footprint_has_changed([self.server])
        self.assertEqual(self.system.total_footprint().value, self.initial_footprint.value)

        logger.warning("Changing back to initial service server")
        self.service.server = self.server
        self.assertEqual(0, new_server.instances_fabrication_footprint.magnitude)
        self.assertEqual(0, new_server.energy_footprint.magnitude)
        self.footprint_has_not_changed([self.server])
        self.assertEqual(self.initial_footprint.value, self.system.total_footprint().value)

    def test_update_storage(self):
        new_storage = Storage(
            "New storage, identical in specs to Default SSD storage",
            carbon_footprint_fabrication=SourceValue(160 * u.kg, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            power=SourceValue(1.3 * u.W, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            lifespan=SourceValue(6 * u.years, Sources.HYPOTHESIS),
            idle_power=SourceValue(0.1 * u.W, Sources.HYPOTHESIS),
            storage_capacity=SourceValue(1 * u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            power_usage_effectiveness=SourceValue(1.2 * u.dimensionless, Sources.HYPOTHESIS),
            average_carbon_intensity=SourceValue(100 * u.g / u.kWh, Sources.HYPOTHESIS),
            data_replication_factor=SourceValue(3 * u.dimensionless, Sources.HYPOTHESIS)
        )
        logger.warning("Changing service storage")
        self.service.storage = new_storage

        self.assertEqual(0, self.storage.instances_fabrication_footprint.magnitude)
        self.assertEqual(0, self.storage.energy_footprint.magnitude)
        self.footprint_has_changed([self.storage])
        self.assertEqual(self.system.total_footprint().value, self.initial_footprint.value)

        logger.warning("Changing back to initial service storage")
        self.service.storage = self.storage
        self.assertEqual(0, new_storage.instances_fabrication_footprint.magnitude)
        self.assertEqual(0, new_storage.energy_footprint.magnitude)
        self.footprint_has_not_changed([self.storage])
        self.assertEqual(self.initial_footprint.value, self.system.total_footprint().value)

    def test_update_uj_steps(self):
        logger.warning("Modifying uj steps")
        new_step = UserJourneyStep(
            "new_step", self.service, SourceValue(5 * u.kB / u.uj),
            SourceValue(5 * u.GB / u.uj),
            user_time_spent=SourceValue(2 * u.min / u.uj), request_duration=SourceValue(4 * u.s))
        self.uj.uj_steps = [new_step]

        self.assertNotEqual(self.initial_footprint.value, self.system.total_footprint().value)
        self.footprint_has_changed([self.storage, self.server, self.network, self.device_population])

        logger.warning("Changing back to previous uj steps")
        self.uj.uj_steps = [self.streaming_step, self.upload_step]

        self.assertEqual(self.initial_footprint.value, self.system.total_footprint().value)
        self.footprint_has_not_changed([self.storage, self.server, self.network, self.device_population])
        self.assertEqual(self.initial_footprint.value, self.system.total_footprint().value)

    def test_update_user_journey(self):
        logger.warning("Changing user journey")
        new_uj = UserJourney("New version of daily Youtube usage", uj_steps=[self.streaming_step])
        self.usage_pattern.user_journey = new_uj

        self.assertNotEqual(self.initial_footprint.value, self.system.total_footprint().value)
        self.footprint_has_changed([self.storage, self.server, self.network, self.device_population])

        logger.warning("Changing back to previous uj")
        self.usage_pattern.user_journey = self.uj

        self.assertEqual(self.initial_footprint.value, self.system.total_footprint().value)
        self.footprint_has_not_changed([self.storage, self.server, self.network, self.device_population])
        self.assertEqual(self.initial_footprint.value, self.system.total_footprint().value)

    def test_update_device_population(self):
        logger.warning("Changing device population")
        new_device_pop = DevicePopulation(
            "New device pop with different specs", SourceValue(10 * u.user), Countries.FRANCE, [default_laptop()])

        self.usage_pattern.device_population = new_device_pop

        self.assertNotEqual(self.initial_footprint.value, self.system.total_footprint().value)
        self.footprint_has_changed([self.storage, self.server, self.network, self.device_population])

        logger.warning("Changing back to initial device population")
        self.usage_pattern.device_population = self.device_population

        self.assertEqual(self.initial_footprint.value, self.system.total_footprint().value)
        self.footprint_has_not_changed([self.storage, self.server, self.network, self.device_population])

    def test_update_country_in_device_pop(self):
        logger.warning("Changing device population country")

        self.device_population.country = Countries.MALAYSIA

        self.assertNotEqual(self.initial_footprint.value, self.system.total_footprint().value)
        self.footprint_has_changed([self.network, self.device_population])

        logger.warning("Changing back to initial device population country")
        self.device_population.country = Countries.FRANCE

        self.assertEqual(self.initial_footprint.value, self.system.total_footprint().value)
        self.footprint_has_not_changed([self.network, self.device_population])

    def test_update_network(self):
        logger.warning("Changing network")
        new_network = Network(
            "New network with same specs as default", SourceValue(0.05 * u("kWh/GB"), Sources.TRAFICOM_STUDY))
        self.usage_pattern.network = new_network

        self.assertEqual(0, self.network.energy_footprint.magnitude)
        self.footprint_has_changed([self.network])
        self.assertEqual(self.system.total_footprint().value, self.initial_footprint.value)

        logger.warning("Changing back to initial network")
        self.usage_pattern.network = self.network
        self.assertEqual(0, new_network.energy_footprint.magnitude)
        self.footprint_has_not_changed([self.network])
        self.assertEqual(self.initial_footprint.value, self.system.total_footprint().value)

    def test_add_uj_step_without_service(self):
        logger.warning("Add uj step without service")

        step_without_service = UserJourneyStep(
            "User checks her phone", None, SourceValue(0 * u.kB / u.uj),
            SourceValue(0 * u.GB / u.uj),
            user_time_spent=SourceValue(20 * u.min / u.uj))

        self.uj.add_step(step_without_service)

        self.footprint_has_not_changed([self.server, self.storage])
        self.footprint_has_changed([self.device_population])
        self.assertNotEqual(self.system.total_footprint().value, self.initial_footprint.value)

        logger.warning("Setting user time spent of the new step to 0s")
        step_without_service.user_time_spent = SourceValue(0 * u.min / u.uj)
        self.footprint_has_not_changed([self.server, self.storage])
        self.assertEqual(self.system.total_footprint().value, self.initial_footprint.value)



from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject

import unittest
from unittest.mock import patch, MagicMock

MODELING_OBJ_CLASS_PATH = "efootprint.abstract_modeling_classes.modeling_object"


class ModelingObjectForTesting(ModelingObject):
    def __init__(self, name, custom_input=None):
        super().__init__(name)
        if custom_input is not None:
            self.custom_input = custom_input

    def compute_calculated_attributes(self):
        pass

    def modeling_objects_whose_attributes_depend_directly_on_me(self):
        return []


class TestModelingObject(unittest.TestCase):

    def setUp(self):
        self.modeling_object = ModelingObjectForTesting("test_object")

    def test_setattr_sets_modeling_obj_container(self):
        value = MagicMock(modeling_obj_container=None)

        with patch(f"{MODELING_OBJ_CLASS_PATH}.type", lambda x: ExplainableObject):
            self.modeling_object.attribute = value

        value.set_modeling_obj_container.assert_called_once_with(self.modeling_object, "attribute")

    def test_handle_model_input_update_triggers(self):
        value = MagicMock(
            modeling_obj_container=None, left_parent=None, right_parent=None, mock_type=ExplainableObject)
        old_value = MagicMock(mock_type=ExplainableObject)
        self.modeling_object.attribute = old_value
        self.modeling_object.handle_model_input_update = MagicMock()

        with patch(f"{MODELING_OBJ_CLASS_PATH}.type", lambda x: x.mock_type):
            self.modeling_object.attribute = value

            self.modeling_object.handle_model_input_update.assert_called_once_with(old_value)

    def test_handle_model_input_update_single_level_descendants(self):
        mod_obj_container = "mod_obj"
        parent = MagicMock()
        parent.id = 'parent_id'

        child1 = MagicMock()
        child1.id = 'child1_id'
        child1.direct_children_with_id = []
        child1.direct_ancestors_with_id = [parent]

        child2 = MagicMock()
        child2.id = 'child2_id'
        child2.direct_children_with_id = []
        child2.direct_ancestors_with_id = [parent]

        for index, child in enumerate([child1, child2]):
            child.modeling_obj_container = mod_obj_container
            child.attr_name_in_mod_obj_container = f"attr_{index}"

        parent.get_all_descendants_with_id.return_value = [child1, child2]
        parent.direct_children_with_id = [child1, child2]

        with patch(f'{MODELING_OBJ_CLASS_PATH}.ModelingObject.retrieve_update_function_from_attribute_name') \
                as mock_retrieve_update_func:
            mock_retrieve_update_func.return_value = MagicMock()

            self.modeling_object.handle_model_input_update(parent)

            mock_retrieve_update_func.assert_any_call(mod_obj_container, "attr_0")
            mock_retrieve_update_func.assert_any_call(mod_obj_container, "attr_1")

            self.assertEqual(mock_retrieve_update_func.call_count, 2)

    def test_handle_model_input_update_multiple_levels_of_descendants(self):
        mod_obj_container = "mod_obj_container"
        parent = MagicMock()
        parent.id = 'parent_id'

        child1 = MagicMock()
        child1.id = 'child1_id'
        child1.direct_ancestors_with_id = [parent]

        grandchild1 = MagicMock()
        grandchild1.id = 'grandchild1_id'
        grandchild1.direct_children_with_id = []
        grandchild1.direct_ancestors_with_id = [child1]

        grandchild2 = MagicMock()
        grandchild2.id = 'grandchild2_id'
        grandchild2.direct_children_with_id = []
        grandchild2.direct_ancestors_with_id = [child1]

        child1.direct_children_with_id = [grandchild1, grandchild2]
        parent.get_all_descendants_with_id.return_value = [child1, grandchild1, grandchild2]
        parent.direct_children_with_id = [child1]

        for index, child in enumerate([child1, grandchild1, grandchild2]):
            child.modeling_obj_container = mod_obj_container
            child.attr_name_in_mod_obj_container = f"attr_{index}"

        with patch(f'{MODELING_OBJ_CLASS_PATH}.ModelingObject.retrieve_update_function_from_attribute_name') \
                as mock_retrieve_update_func:
            mock_retrieve_update_func.return_value = MagicMock()

            self.modeling_object.handle_model_input_update(parent)

            mock_retrieve_update_func.assert_any_call(mod_obj_container, "attr_0")
            mock_retrieve_update_func.assert_any_call(mod_obj_container, "attr_1")
            mock_retrieve_update_func.assert_any_call(mod_obj_container, "attr_2")

            self.assertEqual(mock_retrieve_update_func.call_count, 3)

    def test_launch_attributes_computation_chain(self):
        dep1 = MagicMock()
        dep2 = MagicMock()
        dep1_sub1 = MagicMock()
        dep1_sub2 = MagicMock()
        dep2_sub1 = MagicMock()
        dep2_sub2 = MagicMock()

        self.modeling_object.modeling_objects_whose_attributes_depend_directly_on_me = [dep1, dep2]
        dep1.modeling_objects_whose_attributes_depend_directly_on_me = [dep1_sub1, dep1_sub2]
        dep2.modeling_objects_whose_attributes_depend_directly_on_me = [dep2_sub1, dep2_sub2]

        for obj in [dep1_sub1, dep1_sub2, dep2_sub1, dep2_sub2]:
            obj.modeling_objects_whose_attributes_depend_directly_on_me = []

        self.modeling_object.launch_attributes_computation_chain()

        for dep in [dep1, dep2, dep1_sub1, dep1_sub2, dep2_sub1, dep2_sub2]:
            dep.compute_calculated_attributes.assert_called_once()

    def test_list_attribute_update_works_with_classical_syntax(self):
        val1 = MagicMock()
        val2 = MagicMock()
        val3 = MagicMock()

        mod_obj = ModelingObjectForTesting("test mod obj", custom_input=[val1, val2])

        with patch(f'{MODELING_OBJ_CLASS_PATH}.ModelingObject.handle_object_list_link_update') \
                as mock_list_obj_update_func:
            mod_obj.custom_input = [val1, val2, val3]
            mock_list_obj_update_func.assert_called_once_with([val1, val2, val3], [val1, val2])

    def test_list_attribute_update_works_with_list_condensed_addition_syntax(self):
        val1 = MagicMock()
        val2 = MagicMock()
        val3 = MagicMock()

        mod_obj = ModelingObjectForTesting("test mod obj", custom_input=[val1, val2])

        with patch(f'{MODELING_OBJ_CLASS_PATH}.ModelingObject.handle_object_list_link_update') \
                as mock_list_obj_update_func:
            assert mod_obj.custom_input__previous_list_value_set == [val1, val2]
            mod_obj.custom_input += [val3]
            assert mod_obj.custom_input__previous_list_value_set == [val1, val2, val3]
            mock_list_obj_update_func.assert_called_once_with([val1, val2, val3], [val1, val2])


if __name__ == "__main__":
    unittest.main()
